# LinkText - hyperlink text widget

@class LinkText (Composite)	@file=LinkText

@ The |LinkText| widget is used to display text in multiple fonts
and colors and to provide hyperlinks.  For correct operation, it should
be used only as a child of the accompanying |LinkView| widget, which
provides a scrolling viewport.  This widget also requires the dynamically
growable buffer routines in the file |buffer.c|.


@PUBLIC

@ The |hypertext| resource contains the string that is to be displayed.  
The text may include various formatting tags
which are contained within '|<|' and '|>|' characters.

Font changes are indicated by "|<f?>|" where |?| is a font number from 1 to 5.
The specified font will apply to all text up to the next font change.

Color changes are indicated by "|<c?>|" where ? is a color number from
1 to 8.
The specified color will apply to all text up to the next color change.

A hyperlink is indicated by a string of the form "|<l tag-text>link-text<e>|".
The |link-text| is the text that will appear to the user, highlighted and 
underlined. The |tag-text| is a string that will be passed to a 
client's |activate| callback procedure when the hypertext is clicked.

A dotted tab is a line of dots that extends across the page to a specified
pixel position. It is indicated by "|<d nnn>|" where |nnn| is  a decimal number
specifying the desired pixel position.  If the end position of the dotted tab
is before the current text position, the dotted tab will be placed on the
following line.

A "|<w1>|" tag will cause subsequent text to be wrapped to the width of the
parent viewport widget. Wrapping can be disabled (the default state) with
a "|<w0>|" tag.

A "|<m1>|" tag will cause subsequent text to be horizontally centered. 
Centering can be disabled (the default state) with
a "|<m0>|" tag.

Text can be indented by a specified number of pixels with a "|<i nnn>|" tag,
where |nnn| is  a decimal number
specifying the desired pixel indent.  The indent level remains in effect
until it is explicitly changed, or until a bulleted 
or hanging paragraph is encountered.

A hanging paragraph is where all lines except the first are indented by a
certain number of pixels. A hanging paragraph is indicated by starting the 
paragraph with a "|<h nnn>|" tag, where |nnn| is the number of pixels to
indent. When the paragraph ends (a newline is encountered) the indentation
value is reset to zero.

A bulleted paragraph is indicated by starting the paragraph with a "|<b>|" tag.
The first line of the paragraph will be preceded by a bullet, and subsequent
lines will be indented by the width of the bullet.


	@var String hypertext = NULL

@ The |activate| callback is called when the user clicks on a link. A pointer
to a |LinkRecord| is passed.

	@var <Callback> XtCallbackList activate = NULL

@ The |newpage| callback is called whenever a new page is is first displayed,
either through calling |LinkTextNextPage|, |LinkTextPrevPage|, 
|LinkTextSetPage|, or by setting new hypertext.  A pointer to a 
|PageRecord| is passed which contains the current page and the total number
of pages. The principal use for this
callback is to enable the display of the current page number by a separate
widget.

	@var <Callback> XtCallbackList newpage = NULL

@ The offset from the text to the outside edge of the widgets window.

	@var Dimension internalOffset = 4

@ The text is drawn using the fonts given in the |font1-5| resources. 
The particular font to be used for rendering text 
is specified by font change tags within the
hypertext string.

	@var <FontStruct> XFontStruct *font1 = <String> XtDefaultFont
	@var <FontStruct> XFontStruct *font2 = <String> XtDefaultFont
	@var <FontStruct> XFontStruct *font3 = <String> XtDefaultFont
	@var <FontStruct> XFontStruct *font4 = <String> XtDefaultFont
	@var <FontStruct> XFontStruct *font5 = <String> XtDefaultFont

@ The |color1-8| resources specify the colors used for drawing the text.
The particular color to be used for rendering text
is specified by color change tags within the
hypertext string.

	@var Pixel color1 = <String> XtDefaultForeground
	@var Pixel color2 = <String> XtDefaultForeground
	@var Pixel color3 = <String> XtDefaultForeground
	@var Pixel color4 = <String> XtDefaultForeground
	@var Pixel color5 = <String> XtDefaultForeground
	@var Pixel color6 = <String> XtDefaultForeground
	@var Pixel color7 = <String> XtDefaultForeground
	@var Pixel color8 = <String> XtDefaultForeground

@ The |linkColor| resource specifies the color used for drawing hyperlinks.

	@var Pixel linkColor = <String> "mediumblue"

@ The |selectForeground| and |selectBackground| resources specify the colors 
used for drawing selected text.

	@var Pixel selectForeground = <String> XtDefaultBackground
	@var Pixel selectBackground = <String> XtDefaultForeground

@ Specifies whether hyperlinks should be underlined or not.

	@var Boolean underlineLinks = True

@ The |normalCursor| resource specifies the cursor when not on a link.

	@var Cursor normalCursor = <String> "top_left_arrow"

@ The |linkCursor| resource specifies the cursor when on a link.

	@var Cursor linkCursor = <String> "hand2"

@ The |busyCursor| resource specifies the cursor to be displayed when the 
widget gets told (via the |SetBusy| call) that the application is busy.

	@var Cursor busyCursor = <String> "watch"


@ The |selectThreshold| resource specifies how many pixels the mouse cursor
must be moved before text selection is activated.

	@var int selectThreshold = 5

@ The |spaceWidth| resource is used when converting a text selection into
ASCII text.  For a dotted tab, the pixel measurement is converted to a
number of spaces by dividing by this resource.

	@var int spaceWidth = 5

@ The |indentAscii| resource is used when converting a text selection into
ASCII text.  If set to true (the default) then if the text to be converted
contains a mixture of heading and non-heading lines, then the non-heading
lines will be indented by two spaces in order to make the headings stand out.
If set to false, then no lines will be indented.

	@var Boolean indentAscii = True

@ How big a bullet in a bulleted list should be.

	@var int bulletDiameter = 7

@ How often (in milliseconds) the display will scroll by a line 
when selecting text and the mouse pointer is above or below the viewport.

	@var int scrollInterval = 80


@EXPORTS

@ The |LinkRecord| structure is passed to the clients |activate| callback 
procedure when a hyperlink is clicked on.  The |text| is a string of the text 
actually displayed, |link| is the tag text associated with the hyperlink.

	@type LinkRecord = struct _linkRecord {
		String		text;
		String		link;
		}

	@type PageRecord = struct _pageRecord {
		int		current;
		int		total;
		}



@TRANSLATIONS

@ The following translations allow the page to be grabbed and moved around
inside the parents viewport.

		@trans <Btn2Down>: grabpage()
		@trans <Btn2Motion>: movepage()
		@trans <Btn2Up>: releasepage()
        @trans <Btn1Up>: chaselink() end_selection() 
		@trans <Btn1Down>: start_selection()
		@trans <Btn1Motion>: move_selection()
		@trans <Btn3Down>: extend_selection()

@ The |cursortype| action sets the cursor to either the |linkCursor| or the 
|normalCursor|, depending on whether the mouse pointer is over a hyperlink.

        @trans <Motion>: cursortype()



@PRIVATE

@ This data structure is used to keep the definition of the text that
is to be displayed.

	@type ChunkType = enum { CT_TEXT, CT_LINK, CT_DOTTAB, CT_LINEBREAK, 
							CT_BULLET, CT_TRAVERSER }

	@type TextChunk = struct _textChunk {
		ChunkType	type;
		String		text;
		String		link;
		int			text_attributes;
		int			font;
		int			color;
		int			indent;
		int			pagenum;
		int			x;
		int 		y;
		int			width;
		int 		height;
		int			depth;
		int			chunknum;
		struct _textChunk	*next;
	}


@ A structure which defines the extreme points of a text selection.

	@type SelectionMarker = struct _selectionMarker {
		TextChunk *chunk;
		int chunknum;
		int pos;
		}

@ A structure describing the characteristics of a page when multiple pages
are required.

	@type PageDescription = struct _pageDescription {
		TextChunk *start;
		int height;
		int width;
		}


@ The variable |selecting| is True if the user is currently selecting text with
mouse button 1.

	@var Boolean selecting


@ The variable |selected| is True if this widget holds the current text 
selection.

	@var Boolean selected


@ The variable |asciitext| is a pointer to a dynamically growable buffer 
containing an ASCII translation of the current hypertext.  The translation
is initiated when a text selection transfer is requested, or the
function |HypertextToTex| is called.

	@var Buffer *asciitext

@ The graphics context |ngc| gets modified each time a color changes or a 
font changes in the text.

	@var GC ngc

@ For graying out various things, another GC is used.

	@var GC graygc


@ The variables |head| and |tail| point the extremities of a linked list
of the chunk structures for the text to be displayed in the widget.

	@var TextChunk *head
	@var TextChunk *tail


@ How big the widget would like to be.

	@var int preferred_width
	@var int preferred_height


@ The variable |busy| is true when the widget needs to
display a busy cursor.

	@var Boolean busy


@ The |maxheight| variable contains the largest height of a line. It is used
in the expose routine to know when to stop scanning the list.

	@var int maxheight


@ The mouse position when the user grabs the display with the middle mouse
button.

	@var int grabx
	@var int graby


@ The position of the widget within the parent viewport when the display
was grabbed.

	@var int viewx
	@var int viewy


@ The mouse position in the root window when text selection was started.

	@var int selectx_root
	@var int selecty_root


@ The mouse position in the widget when text selection was started.

	@var int selectx
	@var int selecty


@ The starting and ending positions in the current text selection.

	@var SelectionMarker selectStart
	@var SelectionMarker selectEnd

@ The last selection endpoint which was drawn - used to figure out what
text needs to be redrawn when the selection endpoint is moved.

	@var SelectionMarker lastend

@ The positiion in the root window of the top and bottom borders of the
main viewprt. (Used in scrolling the display when selecting text.)

	@var Position topBorder
	@var Position bottomBorder

@ A timer for implementing scrolling when selecting text

        @var XtIntervalId scrollTimer

@ The |grabCursor| variable contains the cursor to be displayed when
grabbing and moving a page.

	@var Cursor grabCursor 

@ Variables for implementing pages.
	@var int numPages
	@var PageDescription pages[25]
	@var int currentPage
	@var Pixmap pageBitmap



@CLASSVARS
	@var compress_exposure =
		XtExposeCompressMultiple|XtExposeGraphicsExposeMerged



@UTILITIES

@ Add a chunk to the list tail.

@proc add_to_list($, TextChunk *chunk) {
	if($head==NULL) {
		chunk->chunknum=1;
		$head=chunk;
		}
	else {
		chunk->chunknum = $tail->chunknum + 1;
		$tail->next=chunk;
		}

	$tail=chunk;
	}



@ Add a hyperlink.

@proc add_link($, char *txt, int size, int f, char * linkname, int linksize, 
	int attributes, int indent) {
	TextChunk *chunk;

	if (size<0)
		return;

	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_LINK;
	chunk->next=NULL;
	chunk->font=f;
	chunk->color=-1;
	chunk->text_attributes=attributes;
	chunk->indent=indent;

	/* store the chunk text */

	chunk->text=XtMalloc(1+size);
	strncpy(chunk->text,txt, size);
	chunk->text[size]='\0';

	/* store the link text */

	if (linksize<=0)
		chunk->link = chunk->text;
	else {
		chunk->link=XtMalloc(1+linksize);
		strncpy(chunk->link,linkname, linksize);
		chunk->link[linksize]='\0';
		}

	add_to_list($, chunk);
	}



@ Add a page traverser

@proc add_traverser($, TextChunk *prev, TextChunk *next, int pagenum, int y) {
	TextChunk *chunk;

	/* printf("Adding traverser for page %d at position %d\n", pagenum, y); */
	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_TRAVERSER;
	chunk->next=next;
	chunk->pagenum=pagenum;
	chunk->x = 20;
	chunk->y = y;
	chunk->width = trav_width;
	chunk->height = trav_height;
	chunk->depth = 0;
	chunk->text=NULL;
	chunk->link=NULL;

	/* splice into the list */
	if (prev==NULL)
		$head=chunk;
	else
		prev->next = chunk;
	}



@ Add a chunk of text.

@proc add_text($, char *txt, int size, int f, int c, int attributes, int indent) {
	TextChunk *chunk;

	if (size<0)
		return;

	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_TEXT;
	chunk->next=NULL;
	chunk->font=f;
	chunk->color=c;
	chunk->text_attributes=attributes;
	chunk->indent=indent;
	chunk->link=NULL;

	/* store the chunk text */

	chunk->text=XtMalloc(1+size);
	strncpy(chunk->text,txt, size);
	chunk->text[size]='\0';

	add_to_list($, chunk);
	}



@ Add a dotted tab.

@proc add_dottab($, int f, int c, int indent) {
	TextChunk *chunk;

	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_DOTTAB;
	chunk->next=NULL;
	chunk->font=f;
	chunk->color=c;
	chunk->indent=indent;
	chunk->text=NULL;
	chunk->link=NULL;
	chunk->text_attributes=0;

	add_to_list($, chunk);
	}



@ Add a linebreak.

@proc add_linebreak($, int f, int attributes, int indent) {
	TextChunk *chunk;

	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_LINEBREAK;
	chunk->next=NULL;
	chunk->text=NULL;
	chunk->link=NULL;
	chunk->font=f;
	chunk->indent=indent;
	chunk->text_attributes=attributes;

	add_to_list($, chunk);
	}


@ Add a bullet.

@proc add_bullet($, int f, int c) {
	TextChunk *chunk;

	chunk=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk->type=CT_BULLET;
	chunk->next=NULL;
	chunk->text=NULL;
	chunk->link=NULL;
	chunk->indent=0;
	chunk->color=c;
	chunk->font=f;
	chunk->text_attributes=0;

	add_to_list($, chunk);
	}



@PRIVATE

	@type ParseState = enum { PS_TEXT, PS_TAG, PS_FONT, PS_LINK,
		PS_LINKTEXT, PS_LINKEND, PS_DOTTAB, PS_INDENT, PS_HANGINGINDENT,
		PS_COLOR, PS_MIDDLE, PS_WRAP, PS_ENDATTR, PS_BULLET}

@def ATTR_CENTERED 		= 0x01
@def ATTR_WRAPPED  		= 0x02
@def ATTR_HANGINGINDENT = 0x04



@UTILITIES

@ The |EraseLinkText| function is used to erase all of the chunks added
to the text widget.  

@proc EraseLinkText(TextChunk *c) {
	TextChunk *n;
	
	while(c) {
		n=c->next;
		if ((c->link != NULL) && (c->link != c->text))
			XtFree((char *)c->link);
		XtFree((char *)c->text);
		XtFree((char *)c);
		c=n;
		}
	}



@ Convert a hypertext string into chunks.

@proc hyperchunkit($) {
	char *cp, *start, *end;
	char *linkstart, *linkend;
	int font=1, newfont, color=1, newcolor;
	int newindent, indent=0;
	ParseState cstate=PS_TEXT;
	int dottab;
	int attributes=0;
	int new_attributes;
	Boolean inbullet=False;
		
	/* printf("Hyperchunking\n"); */
	assert(XtIsSubclass($, linkTextWidgetClass));

	EraseLinkText($head);
	$head=$tail=NULL;


	cp=$hypertext;
	start=end=cp;
	linkstart=linkend=NULL;

	while (*cp != '\0') {
		switch (cstate) {
			case PS_DOTTAB:
				if (*cp=='>') {
					add_dottab($, font, color, dottab);
					start=cp+1;
					cstate=PS_TEXT;
					}
				else if (isdigit(*cp))
					dottab = dottab*10 + (*cp-'0');
				else if (*cp != ' ')
					cstate=PS_TEXT;
				break;

			case PS_INDENT:
				if (*cp=='>') {
					indent=newindent;
					start=cp+1;
					cstate=PS_TEXT;
					}
				else if (isdigit(*cp))
					newindent = newindent*10 + (*cp-'0');
				else if (*cp != ' ')
					cstate=PS_TEXT;
				break;


			case PS_HANGINGINDENT:
				if (*cp=='>') {
					indent=newindent;
					attributes |= ATTR_HANGINGINDENT;
					start=cp+1;
					cstate=PS_TEXT;
					}
				else if (isdigit(*cp))
					newindent = newindent*10 + (*cp-'0');
				else if (*cp != ' ')
					cstate=PS_TEXT;
				break;


			case PS_ENDATTR:
				if (*cp=='>') {
					attributes=new_attributes;
					start=cp+1;
					}
				cstate=PS_TEXT;
				break;
			case PS_MIDDLE:
				if (*cp=='1') {
					new_attributes = attributes | ATTR_CENTERED; 
					cstate=PS_ENDATTR;
					}
				else if (*cp=='0') {
					new_attributes = attributes & ~ATTR_CENTERED; 
					cstate=PS_ENDATTR;
					}
				else
					cstate=PS_TEXT;
				break;

			case PS_WRAP:
				if (*cp=='1') {
					new_attributes = attributes | ATTR_WRAPPED; 
					cstate=PS_ENDATTR;
					}
				else if (*cp=='0') {
					new_attributes = attributes & ~ATTR_WRAPPED; 
					cstate=PS_ENDATTR;
					}
				else
					cstate=PS_TEXT;
				break;


			case PS_BULLET:
				if (*cp=='>') {
					add_bullet($, font, color);
					indent=3*$bulletDiameter;
					inbullet=True;
					start=cp+1;
					}
				cstate=PS_TEXT;
				break;

			case PS_LINKEND:
				if (*cp=='>') {
					start=cp+1;
					cstate=PS_TEXT;
					}
				break;


			case PS_LINK:
				if (*cp=='>') {
					linkend=cp;
					cstate=PS_LINKTEXT;
					start=cp+1;
					}
				break;


			case PS_FONT:
				if (*cp=='>') {
					font=newfont;
					start=cp+1;
					cstate=PS_TEXT;
					}
				else if (isdigit(*cp))
					newfont = newfont*10 + (*cp-'0');
				else if (*cp != ' ')
					cstate=PS_TEXT;
				break;


			case PS_COLOR:
				if (*cp=='>') {
					color=newcolor;
					start=cp+1;
					cstate=PS_TEXT;
					}
				else if (isdigit(*cp))
					newcolor = newcolor*10 + (*cp-'0');
				else if (*cp != ' ')
					cstate=PS_TEXT;
				break;


			case PS_TAG:
				switch (*cp) {
					case 'b':	cstate=PS_BULLET; break;
					case 'w':	cstate=PS_WRAP; break;
					case 'm':	cstate=PS_MIDDLE; break;
					case 'c':	cstate=PS_COLOR; newcolor=0; break;
					case 'd':	cstate=PS_DOTTAB; dottab=0; break;
					case 'i':	cstate=PS_INDENT; newindent=0; break;
					case 'h':	cstate=PS_HANGINGINDENT; newindent=0; break;
					case 'f':	cstate=PS_FONT; newfont=0; break;
					case 'l':	if (*(cp+1)==' ') {
									cstate=PS_LINK; 
									linkstart=cp+2; 
									}
								break;
					default:	cstate=PS_TEXT;
					}
				break;


			case PS_LINKTEXT:
				if (*cp=='\n') {
					/* for when a newline appears in the middle of a link */
					add_link($, start, cp-start, font, linkstart, 
						linkend-linkstart, attributes, indent);
					if (inbullet) {
						inbullet=False;
						indent=0;
						}
					add_linebreak($, font, attributes, indent);
					if (attributes | ATTR_HANGINGINDENT) {
						attributes &= ~ATTR_HANGINGINDENT;
						indent=0;
						}
					start=cp+1;
					}
				else if (*cp=='<') {
					add_link($, start, cp-start, font, linkstart, 
						linkend-linkstart, attributes, indent);
					start=cp;
					cstate=PS_LINKEND;
					}
				break;


			case PS_TEXT:
				if (*cp=='\n') {
					if (cp!=start)
						add_text($, start, cp-start, font, color, 
							attributes, indent);
					if (inbullet) {
						inbullet=False;
						indent=0;
						}
					if (attributes | ATTR_HANGINGINDENT) {
						attributes &= ~ATTR_HANGINGINDENT;
						indent=0;
						}
					add_linebreak($, font, attributes, indent);
					start=cp+1;
					break;
					}
				else if (*cp=='<') {
					if (cp > start)
						add_text($, start, cp-start, font, color, 
							attributes, indent);
					start=cp;
					cstate=PS_TAG;
					}
				break;
			

			default:
				XtAppWarning(XtWidgetToApplicationContext($), 
					"hyperchunkit: unknown state!");
			}
		cp++;	
		}

	if ((cstate==PS_TEXT) && (cp > start))
		add_text($, start, cp-start, font, color, attributes, indent);


	}



@ Create the GC used for drawing text.

@proc make_gc($) {
	XtGCMask mask, dynmask, dontcaremask;
	XGCValues values;

	values.foreground=$color1;
	values.background=$background_pixel;
	values.font=$font1->fid;
	dynmask = GCForeground | GCBackground | GCFont;
	dontcaremask = GCJoinStyle | GCFillRule | GCTile
		| GCTileStipXOrigin | GCTileStipYOrigin | GCDashOffset | GCDashList
		| GCStipple | GCGraphicsExposures | GCArcMode | GCLineWidth | GCClipMask;
	mask=0;
	$ngc=XtAllocateGC($,$depth,mask,&values,dynmask,dontcaremask);
	}	

@ The |make_graygc| routine creates a GC for graying out the text. It
contains a stipple in the background color, that will be applied over
the text.

@proc make_graygc($) {
    XtGCMask mask;
    XGCValues values;

    if ($graygc != NULL) XtReleaseGC($, $graygc);
    values.foreground = $background_pixel;
    values.stipple =
        XCreateBitmapFromData(XtDisplay($),
                              RootWindowOfScreen(XtScreen($)),
                              stip4_bits, stip4_width, stip4_height);
    values.fill_style = FillStippled;
    mask = GCForeground | GCStipple | GCFillStyle;
    $graygc = XtGetGC($, mask, &values);
	}


@ This function returns |True| if the string ends with a newline.

@proc int newline(char *s) {
	if (s == NULL)
		return 0;
	return (s[strlen(s)-1] == '\n');
	}



@ This function is used to calculate the maximum ascender and descender
for a line of text chunks.  This could be the ascent from one font and
the descent from a different font to give the maximum line height.

@proc compute_line_height($, TextChunk *chunk, int *ma, int *md) {
	XFontStruct *f;
	char *s;

	*ma=0, *md=0;
	if(chunk==NULL) return;
	do {
		switch(chunk->font)
		{
			case 2:		f=$font2;	break;
			case 3:		f=$font3;	break;
			case 4:		f=$font4;	break;
			case 5:		f=$font5;	break;
			default:	f=$font1;	break;
		}

		*ma = *ma>f->ascent ? *ma : f->ascent;
		*md = *md>f->descent ? *md : f->descent;
		s = chunk->text;
 		chunk = chunk->next;
		} 
	while (chunk && !newline(s) && (chunk->type!=CT_LINEBREAK));
	}



@ Set the graphics context to draw in the appropriate color.

@proc set_color($, int c, Boolean reverse) {
	if (reverse) {
		XSetForeground(XtDisplay($), $ngc, $selectForeground);
		XSetBackground(XtDisplay($), $ngc, $selectBackground);
		}
	else {
		XSetBackground(XtDisplay($), $ngc, $background_pixel);
		switch(c) {
			case -1:	
				XSetForeground(XtDisplay($), $ngc, $linkColor);
				break;
			case 2:	XSetForeground(XtDisplay($), $ngc, $color2);
				break;
			case 3:	XSetForeground(XtDisplay($), $ngc, $color3);
				break;
			case 4:	XSetForeground(XtDisplay($), $ngc, $color4);
				break;
			case 5:	XSetForeground(XtDisplay($), $ngc, $color5);
				break;
			case 6:	XSetForeground(XtDisplay($), $ngc, $color6);
				break;
			case 7:	XSetForeground(XtDisplay($), $ngc, $color7);
				break;
			case 8:	XSetForeground(XtDisplay($), $ngc, $color8);
				break;
			default:XSetForeground(XtDisplay($), $ngc, $color1);
				break;
			}
		}
	}



@ The |draw_string| utility function is used to draw a string at a set
position, color and font in the widget.

@proc draw_string($, XFontStruct *f, TextChunk *c, char *s, int length,
	int x, int y, Boolean reverse) {
	int w;

	XSetFont(XtDisplay($), $ngc, f->fid);
	set_color($, c->color, reverse);			
	XDrawImageString(XtDisplay($),XtWindow($),$ngc,
			x,y,s,length);


	/* draw the underline for links */
	if ((c->link!=NULL) && $underlineLinks) {
		w = XTextWidth(f,s,length);
		XDrawLine(XtDisplay($),XtWindow($),$ngc,
			x, y+1, x+w, y+1);
		}
	}



@ The |draw_chunktext| utility function is used to draw the text for
a given chunk at a set position, color and font in the widget.  Some part
of the text may need to be drawn in reverse video to indicate a text
selection.

@proc draw_chunktext($, XFontStruct *f, TextChunk *c, char *s, int length, int x, int y) {
	SelectionMarker top, bottom;
	int offs;

	/* if nothing is selected then just draw it */
	if (!$selected) {
		draw_string($, f, c, s, length, x, y, False);
		return;
		}

	/* compute the top and bottom point of the text selection */
	top=minimum_marker($selectStart, $selectEnd);
	bottom=maximum_marker($selectStart, $selectEnd);

	/* are we above or below the affected chunks? */
	if ((c->chunknum < top.chunknum) || (c->chunknum > bottom.chunknum)) {
		draw_string($, f, c, s, length, x, y, False);
		return;
		}

	if (top.chunknum==bottom.chunknum) {
		offs=0;
		if (top.pos>0) {
			draw_string($, f, c, s, top.pos, x, y, False);
			offs += XTextWidth(f, s, top.pos);
			}
		if (bottom.pos > top.pos) {
			draw_string($, f, c, s+top.pos, bottom.pos-top.pos, x+offs, y, True);
			offs += XTextWidth(f, s+top.pos, bottom.pos-top.pos);
			}
		if (bottom.pos<length)
			draw_string($, f, c, s+bottom.pos, length-bottom.pos+1, x+offs, y, False);
		}
	else if (c->chunknum==top.chunknum) {
		offs=0;
		if (top.pos>0) {
			draw_string($, f, c, s, top.pos, x, y, False);
			offs += XTextWidth(f, s, top.pos);
			}
		if (top.pos<length)
			draw_string($, f, c, s+top.pos, length-top.pos, x+offs, y, True);
		}
	else if (c->chunknum==bottom.chunknum) {
		offs=0;
		if (bottom.pos>0) {
			draw_string($, f, c, s, bottom.pos, x, y, True);
			offs += XTextWidth(f, s, bottom.pos);
			}
		if (bottom.pos<length)
			draw_string($, f, c, s+bottom.pos, length-bottom.pos+1, x+offs, y, False);
		}
	else
		draw_string($, f, c, s, length, x, y, True);
	}



@ Remove all page breaks.

@proc remove_page_breaks($) {
	TextChunk *chunk, *prev;

	if ($head==NULL)
		return;

	chunk=$head;
	if ($head->type == CT_TRAVERSER) {
		$head=$head->next;
		XtFree((char*) chunk);
		chunk=$head;
		}

	prev=$head;
	while (chunk!=NULL) {
		if (chunk->type == CT_TRAVERSER) {
			prev->next = chunk->next;
			XtFree((char*) chunk);
			chunk = prev->next;
			}
		else {
			prev=chunk;
			chunk=chunk->next;
			}
		}
	}


@ This procedure is simply for debugging purposes.

@proc dump_positions($) {
	TextChunk *chunk;

	for(chunk=$head ; chunk ; chunk=chunk->next) {
		printf("%3d: ", chunk->chunknum);
		switch (chunk->type) {
			case CT_TRAVERSER:
				printf("Page traverser - page=%d: (%d, %d) h=%d, d=%d, w=%d\n",
					chunk->pagenum, chunk->x, chunk->y, chunk->height, chunk->depth, chunk->width);
				break;
			case CT_BULLET:
				printf("Bullet: (%d, %d) h=%d, d=%d, w=%d\n",
					chunk->x, chunk->y, chunk->height, chunk->depth, chunk->width);
				break;
			case CT_LINEBREAK:
				printf("Linebreak: (%d, %d) h=%d, d=%d, w=%d\n",
					chunk->x, chunk->y, chunk->height, chunk->depth, chunk->width);
				break;
			case CT_DOTTAB:
				printf("Dottab: (%d, %d) h=%d, d=%d, w=%d\n",
					chunk->x, chunk->y, chunk->height, chunk->depth, chunk->width);
				break;
			case CT_TEXT:
				printf("Text: (%d, %d) h=%d, d=%d, w=%d, attr=%d, indent=%d, text=\"%s\"\n",
					chunk->x, chunk->y, chunk->height, chunk->depth, 
					chunk->width, chunk->text_attributes, chunk->indent, chunk->text);
				break;
			case CT_LINK:
				printf("Link: (%d, %d) h=%d, d=%d, w=%d, attr=%d, indent=%d, text=%s, link=%s\n",
					chunk->x, chunk->y, chunk->height, chunk->depth, 
					chunk->width, chunk->text_attributes, chunk->indent, chunk->text, 
					chunk->link);
				break;
			default:
				XtAppWarning(XtWidgetToApplicationContext($), 
					"dump_positions() - unknown type!");
			}
		}
	}



@ Split the given text (or link) chunk into two pieces. The text of the second
chunk will start at position |p|.

@proc split_chunk($, TextChunk *chunk, int p) {
	TextChunk *chunk2;
	char *c = chunk->text + p;

	/* sanity check */
	if ((p==0) || (p>strlen(chunk->text)))
		return;

	chunk2=(TextChunk*)XtMalloc(sizeof(TextChunk));
	chunk2->type=chunk->type;
	chunk2->next=chunk->next;
	chunk2->font=chunk->font;
	chunk2->color=chunk->color;
	chunk2->text_attributes=chunk->text_attributes;
	chunk2->indent=chunk->indent;
	if (chunk->link != NULL)
		chunk2->link=XtNewString(chunk->link);
	else
		chunk2->link=NULL;
	chunk2->text=XtNewString(c);
	*c='\0';
	c=XtNewString(chunk->text);
	XtFree(chunk->text);
	chunk->text=c;
	chunk->next=chunk2;
	}



@ Break the given chunk of text into two chunks such that the width of the
first chunk plus |x| is less than |width|. 

@proc compute_linebreaks($, TextChunk *chunk, XFontStruct *f, int *x, 
	int *y, int width) {
	char *c, *n;
	int w;
	Boolean atomic=False;

	c=NULL;
	n=strchr(chunk->text, ' ');
	if (n==NULL)
		atomic=True;
	else
		atomic= (*(n+1)=='\0');

	/* if the chunk is unbreakable, insert a soft line break only if we
	   are not at the start of a new line */
	if (atomic)  {
		if (*x>$internalOffset+chunk->indent) {
			*x=$internalOffset+chunk->indent;
			*y += f->descent + f->ascent;
			}
		pin_chunk($, chunk, f, x, y);
		return;
		}

	/* if the first unit in the chunk doesn't fit then insert a soft line
	   break only if we are not at the start of a new line */
	w=XTextWidth(f, chunk->text, n - chunk->text +1);
	if (*x + w > width) {
		split_chunk($, chunk, n-chunk->text+1);
		if (*x>$internalOffset+chunk->indent) {
			*x=$internalOffset+chunk->indent;
			*y += f->descent + f->ascent;
			}
		pin_chunk($, chunk, f, x, y);
		return;
		}

	c=n;
	n=strchr(n+1, ' ');
	while (n!=NULL) {
		w=XTextWidth(f, chunk->text, n - chunk->text +1);
		if (*x+w >= width) 
			break;
		c=n;
		n=strchr(n+1, ' ');
		}
	split_chunk($, chunk, c-chunk->text+1);
	pin_chunk($, chunk, f, x, y);
	*x=$internalOffset+chunk->indent;
	*y += f->descent + f->ascent;
	}


@ For the given chunk, set its internal variables which specify its position
and size.

@proc pin_chunk($, TextChunk *chunk, XFontStruct *f, int *x, int *y) {
	int w;

	w = XTextWidth(f,chunk->text,strlen(chunk->text));
	chunk->width = w;
	chunk->height = f->ascent;
	chunk->depth = f->descent;
	chunk->x = *x;
	chunk->y = *y;
	*x += chunk->width;
	}



@ Calculate the position and size of each chunk of text and the overall 
size of the widget.

@proc compute_positions($, int preferred_width) {
	TextChunk *chunk, *prev;
	TextChunk *chunkc;
	int x, y, w, n;
	int endx, coffset, cy;
	int maxw=0;
	int ma, md;
	XFontStruct *f;
	char *s;
	int tabpos;
	Boolean inhanging=False;
	Boolean firstline;
	int page, yoffset;
	static PageRecord pr;

	/* remove all page breaks */
	remove_page_breaks($);

	/* initial position of text */
	compute_line_height($, $head, &ma, &md);
	y = $internalOffset + ma;
	x = $internalOffset;
	
	/* printf("%s", $hypertext); */

	/* consider each chunk of text in turn and work out where it will be put */
	for (chunk=$head ; chunk ; chunk=chunk->next) {

		switch(chunk->font)	{
			case 2:		f=$font2;	break;
			case 3:		f=$font3;	break;
			case 4:		f=$font4;	break;
			case 5:		f=$font5;	break;
			default:	f=$font1;	break;
			}

		switch (chunk->type) {
			case CT_BULLET:
				chunk->x = x;
				chunk->y = y;
				chunk->width=3 * $bulletDiameter;
				chunk->depth=f->descent;
				chunk->height=f->ascent; /*$bulletDiameter;*/
				x += chunk->width;
				break;
			case CT_LINEBREAK:
				chunk->x = x;
				chunk->y = y;
				chunk->width=1000;	/* used in text selections */
				chunk->depth=0;
				chunk->height=10;
				inhanging=False;
				break;
			case CT_DOTTAB:
				if (chunk->indent > 0) {
						tabpos = chunk->indent + $internalOffset;
					if (tabpos <= x) {
						/* check for a line width maxima */
						maxw = maxw>x ? maxw : x;
						x = tabpos-(chunk->indent/2); 
						if (x<$internalOffset)
							x=$internalOffset;
						compute_line_height($, chunk->next, &ma, &md);
						y += ma + md;
						}
					/* record the chunk co-ordinates */
					chunk->x = x;
					chunk->y = y;
					chunk->height = 5;   /* a guess is good enough here ? */
					chunk->depth = 0;
					chunk->width = tabpos - x;
					x = tabpos;
					}
				break;
			case CT_TEXT:
			case CT_LINK:
				s=chunk->text;
				if (s==NULL)
					XtError("compute_positions: null text");
				if (*s=='\0')
					XtError("compute_positions: empty text");
				
				if ((chunk->text_attributes & ATTR_HANGINGINDENT) && !inhanging)
					firstline=True;
				inhanging=(chunk->text_attributes & ATTR_HANGINGINDENT);
				if (x<chunk->indent+$internalOffset) 
					if ((inhanging && !firstline) || (!inhanging))
						x=chunk->indent+$internalOffset;

				/* compute the width */
				w = XTextWidth(f,s,strlen(s));
				if ((chunk->text_attributes & ATTR_WRAPPED) &&
						(x+w > preferred_width-$internalOffset)) {
					compute_linebreaks($, chunk, f, &x, &y, preferred_width-$internalOffset);
					firstline=False;
					}
				else
					pin_chunk($, chunk, f, &x, &y);
				break;

			default:
				XtAppWarning(XtWidgetToApplicationContext($), 
					"compute_position: unknown chunk!");
			}

		/* if we're moving to a new line, check if the current line
		   is the widest we've seen so far */
		if (newline(chunk->text) || (chunk->type==CT_LINEBREAK)) {
			maxw = maxw>x ? maxw : x;
			x=$internalOffset;
			y += md;
			compute_line_height($, chunk->next, &ma, &md);
			$maxheight = $maxheight> ma ? $maxheight : ma;
			y += ma;
		}
	}

	$pages[0].height= y + md + $internalOffset;

	/* paginate if height exceeds maximum window height */
	$pages[0].start=$head;
	$numPages=1;
	$currentPage=0;
	if ($pages[0].height > 32767) {
		yoffset=trav_height+30;
		page=0;
		chunk=$head;
		add_traverser($, NULL, $head, page, 10+trav_height);
		chunk = $head->next;
		prev=$head;
		$pages[0].start=$head;
		for ( ; chunk ; prev=chunk, chunk=chunk->next) {
			chunk->y += yoffset;
			if (chunk->y > 32000) {
				add_traverser($, prev, chunk, page, prev->y+20+trav_height);
				$pages[page].height=prev->y+30+trav_height;
				prev=prev->next;
				page++;
				chunk->y -= yoffset;  	/* restore original y pos */
				yoffset=trav_height+30 - chunk->y;
				chunk->y += yoffset;
				add_traverser($, prev, chunk, page, 10+trav_height);
				prev=prev->next;
				$pages[page].start=prev;
				}
			}	
		add_traverser($, prev, NULL, page, prev->y+20+trav_height);
		$numPages=page+1;
		$pages[page].height=prev->y+30+trav_height;
		}

	$height=$preferred_height = $pages[0].height;

	/* scan through the list:
		- recompute chunknums
		- compute maximum width
		- compute maximum line height
	 */

	n=1;
	$maxheight=0;
	page=0;
	$pages[0].width=0;
	for (chunk=$head ; chunk ; chunk=chunk->next) {
		chunk->chunknum = n++;
		if (chunk->type==CT_TRAVERSER) {
			if (page!=chunk->pagenum) {
				page=chunk->pagenum;
				$pages[page].width=$internalOffset*2 + trav_width;
				}
			}

		if (chunk->height > $maxheight)
			$maxheight = chunk->height;
		if (chunk->type != CT_LINEBREAK)
			if (chunk->x + chunk->width + $internalOffset > $pages[page].width)
				$pages[page].width = chunk->x + chunk->width + $internalOffset;
		}

	$width=$preferred_width =  $pages[0].width;

	/* scan through the list and center lines	*/

	for(chunk=$head ; chunk ; chunk=chunk->next) {
		if (!(chunk->text_attributes&ATTR_CENTERED))
			continue;
		if (chunk->x!=$internalOffset)
			continue;

		/* find out how wide the line is */
		chunkc=chunk;
		cy=chunkc->y;
		endx=0;
		while (chunkc && (chunkc->text_attributes&ATTR_CENTERED) && (chunkc->y == cy)) {
			if (chunkc->type != CT_LINEBREAK)
				endx=chunkc->x + chunkc->width;
			chunkc=chunkc->next;
			}

		/* now adjust the x positions */
		chunkc=chunk;
		coffset=(preferred_width-$internalOffset-endx)/2;
		if (coffset>0)
			while (chunkc && (chunkc->text_attributes&ATTR_CENTERED) && (chunkc->y == cy)) {
				chunkc->x += coffset;
				chunkc=chunkc->next;
				}
		}

	/* dump_positions($); */
	pr.current=$currentPage;
	pr.total=$numPages;
	XtCallCallbackList($, $newpage, &pr);
	}





@METHODS

@ The |initialize| method sets the private variables to default values.
The strange call to |XmuInternAtom| is necessary to initialize the Xmu
atom caching mechanism. |XA_TARGETS| is used by the selection procedures.

@proc initialize {
	Pixmap shape, shapemask;
	XColor fore, back;

	$busy=False;
	$selecting=False;
	$selected=False;
	$selectStart.chunknum=-1;  /* nothing selected */
	$lastend.chunknum=-1; 
	$head=NULL;
	$tail=NULL;
	$asciitext=NULL;
	(void) XmuInternAtom(XtDisplay($), XmuMakeAtom("NULL"));

    if ($hypertext) {
		$hypertext = XtNewString($hypertext);
		hyperchunkit($);
		}

	$scrollTimer=0;

	$ngc=NULL;
	make_gc($);
	$graygc=NULL;
	make_graygc($);

	/* make the grabCursor */
	shape=XCreatePixmapFromBitmapData(XtDisplay($), 
		RootWindowOfScreen(XtScreen($)),
		hand_bits, hand_width, hand_height, 1, 0, 1);
	shapemask=XCreatePixmapFromBitmapData(XtDisplay($), 
		RootWindowOfScreen(XtScreen($)),
		handmask_bits, handmask_width, handmask_height, 1, 0, 1);

	XParseColor(XtDisplay($), $colormap, "black", &fore);
	XParseColor(XtDisplay($), $colormap, "white", &back);
	$grabCursor=XCreatePixmapCursor(XtDisplay($), shape, shapemask, 
		&fore, &back, 0, 0);

	}



@proc realize {
    #realize($, mask, attributes);
    $pageBitmap = XCreateBitmapFromData(XtDisplay($), XtWindow($),
        trav_bits, trav_width, trav_height);
	}



@ The |destroy| method releases memory and resources held by the widget.

@proc destroy {
	XtReleaseGC($, $ngc);
	EraseLinkText($head);
	$head=$tail=NULL;
}



@ The |set_values| method copies all changed string resources.

@proc set_values {
	Boolean do_size = False;
	Boolean do_redisplay = False;
	Dimension preferred_width = 500;

	/* printf("Set Values:\n"); */

	XtVaGetValues(XtParent($), XtNwidth, &preferred_width, NULL);
	preferred_width -= 16;    /* KLUDGE - adjust for scrollbar */

	if ($background_pixel != $old$background_pixel)
		make_graygc($);

    if ($hypertext != $old$hypertext) {
        XtFree($old$hypertext);
        $hypertext = XtNewString($hypertext);
		hyperchunkit($);
		do_size = True;
		$selected=False;
		$selectStart.chunknum=-1;
		}

	if (($old$internalOffset != $internalOffset) ||
		($old$width != $width)) {
		do_size = True;
		$selected=False;
		preferred_width=$width;
		}

	if (($old$font1 != $font1) ||
		($old$font2 != $font2) ||
		($old$font3 != $font3) ||
		($old$font4 != $font4) ||
		($old$font5 != $font5))
		do_size = True;

	if (($old$color1 != $color1) ||
		($old$color2 != $color2) ||
		($old$color3 != $color3) ||
		($old$color4 != $color4) ||
		($old$color5 != $color5) ||
		($old$color6 != $color6) ||
		($old$color7 != $color7) ||
		($old$color8 != $color8) ||
		($old$linkColor != $linkColor))
		do_redisplay = True;
	
	if ($selected)
		if (($old$selectBackground != $selectBackground) ||
			($old$selectForeground != $selectForeground))
			do_redisplay = True;

	if ($old$underlineLinks != $underlineLinks)
		do_redisplay=True;
		
	if (do_size) {
		compute_positions($, preferred_width);
		do_redisplay=True;
		}

	return do_redisplay;
	}



@ The |query_geometry| method is called when the parent wants to resize
the widget.
	
@proc query_geometry {
	/* say no to any requested size change */
	reply->request_mode = CWWidth |CWHeight;
	reply->width = $preferred_width;
	reply->height = $preferred_height;
	return XtGeometryAlmost;
	}


@ The |expose| method is called when any part of the widget needs to be drawn.
	
@proc expose {
	TextChunk *chunk;
	int x, y, height, depth, width;
	int x1, cw;
	XFontStruct *f;
	char *s;
	char dots[256];
	int ndots, dotwidth, i, slen;
	int exp_y1, exp_y2;
	SelectionMarker top, bottom;
	Boolean reverse;
	int last_height=$font1->ascent;
	int last_depth=$font1->descent;

		
	if(! XtIsRealized($))
		return ;

	if (event == NULL) {
		exp_y1=0;
		exp_y2=$height;
		}
	else {
		exp_y1=event->xexpose.y;
		exp_y2= exp_y1+event->xexpose.height;
		}

	/* don't try to draw to portions of the window that are off screen */
	if (-$y > exp_y1)
		exp_y1 = -$y;

	if (exp_y2 > exp_y1 +1500)
		exp_y2 = exp_y1 + 1500;

	/* printf("expose %d to %d, y=%d\n", exp_y1, exp_y2, $y); */

	/* skip past any chunks that are 'above' the region to be
			redisplayed */

	for(chunk=$pages[$currentPage].start ; chunk ; chunk=chunk->next) {
		if (chunk->y + chunk->depth > exp_y1)
			break;
		switch(chunk->font)	{
			case 2:		f=$font2;	break;
			case 3:		f=$font3;	break;
			case 4:		f=$font4;	break;
			case 5:		f=$font5;	break;
			default:	f=$font1;	break;
			}
		if (chunk->type == CT_LINEBREAK)
			if (chunk->next != NULL)
				last_height = chunk->next->y - chunk->y - f->descent;
		}
		
	/* compute the top and bottom point of the text selection */
	top=minimum_marker($selectStart, $selectEnd);
	bottom=maximum_marker($selectStart, $selectEnd);

	for( ; chunk; chunk=chunk->next) 	{
/* printf("type %d at (%d, %d)\n", chunk->type, chunk->x, chunk->y);*/
		switch(chunk->font)	{
			case 2:		f=$font2;	break;
			case 3:		f=$font3;	break;
			case 4:		f=$font4;	break;
			case 5:		f=$font5;	break;
			default:	f=$font1;	break;
			}

		x=chunk->x;
		y=chunk->y;
		height=chunk->height;
		depth=chunk->depth;
		width=chunk->width;

		/* see if remaining text is below the exposed area */
		if (y - $maxheight > (exp_y2))
			break;

		/* see if we've reached the end of the page */
		if (chunk->type == CT_TRAVERSER)
			if (chunk->pagenum != $currentPage)
				break;
		switch (chunk->type) {
			case CT_LINK:
			case CT_TEXT:
				s=chunk->text;
				x1=x;
				slen = strlen(s);
				if (newline(s))
					slen--;
				cw=XTextWidth(f,s,slen);
				draw_chunktext($, f, chunk, s, slen, x1, y);
				last_height=f->ascent;
				last_depth=f->descent;
				break;


			case CT_DOTTAB:		
				if (chunk->indent > 0) {
					dotwidth = XTextWidth(f,".",1);
					ndots = (chunk->indent -x)/dotwidth;
					for (i=0; i<ndots; i++)
						dots[i]='.';
					 draw_chunktext($, f, chunk, dots, ndots, x, y);
					}
				break;

			case CT_BULLET:
				reverse=$selected &&
						(chunk->chunknum >= top.chunknum) &&
						(chunk->chunknum <= bottom.chunknum);
				if (reverse) {
					XSetForeground(XtDisplay($), $ngc, $selectBackground);
					XSetBackground(XtDisplay($), $ngc, $selectForeground);
					XFillRectangle(XtDisplay($), XtWindow($), $ngc, 
						chunk->x, chunk->y-chunk->height, 
						chunk->width, chunk->height+chunk->depth);
					}
				else
					XClearArea(XtDisplay($), XtWindow($), 
						chunk->x, chunk->y-chunk->height, 
						chunk->width, chunk->height+chunk->depth, False);
				set_color($, chunk->color, reverse);
				XFillArc(XtDisplay($), XtWindow($), $ngc, 
					chunk->x + $bulletDiameter, chunk->y - $bulletDiameter,
					$bulletDiameter, $bulletDiameter, 0, 360*64);
				break;

			case CT_LINEBREAK:
				reverse=$selected &&
						(chunk->chunknum >= top.chunknum) &&
						(chunk->chunknum <= bottom.chunknum);
				if (reverse) {
					XSetForeground(XtDisplay($), $ngc, $selectBackground);
					XSetBackground(XtDisplay($), $ngc, $selectForeground);
					}
				else {
					XSetForeground(XtDisplay($), $ngc, $background_pixel);
					XSetBackground(XtDisplay($), $ngc, $selectBackground);
					}
				XFillRectangle(XtDisplay($), XtWindow($), $ngc, 
						chunk->x, chunk->y-last_height, 
						10, last_height+last_depth);
				if (chunk->next != NULL)
					last_height = chunk->next->y - chunk->y - f->descent;
				break;

			case CT_TRAVERSER:
				/* printf("Drawing traverser at (%d, %d)\n", chunk->x, chunk->y); */
				XSetForeground(XtDisplay($), $ngc, $linkColor);
				XSetBackground(XtDisplay($), $ngc, $background_pixel);
				XCopyPlane(XtDisplay($), $pageBitmap, XtWindow($), $ngc,
					0, 0, trav_width, trav_height, chunk->x, chunk->y-trav_height, 1);

				/* grey out inoperative buttons */
				if (chunk->pagenum == 0) {
					XFillRectangle(XtDisplay($), XtWindow($), $graygc, 
						chunk->x+2, chunk->y-trav_height+2, trav_width/2-4, trav_height-4);
					}
				else if (chunk->pagenum == $numPages-1)
					XFillRectangle(XtDisplay($), XtWindow($), $graygc, 
						chunk->x+trav_width/2+2, chunk->y-trav_height+2, 
						trav_width/2-4, trav_height-4);
				break;

			default:
				XtAppWarning(XtWidgetToApplicationContext($), 
					"expose() - unknown type!");
			}
		
		}
	}



@UTILITIES

@ Given a mouse position |(x,y)| this procedure computes which chunk of text
and which character position within the chunk that the mouse position
corresponds to.

@proc find_position($, int x, int y, SelectionMarker *marker) {
	TextChunk *c, *prev;
	int cury, i, xpos, slen;
	XFontStruct *f;
	
	if ($head==NULL)
		return;

	(*marker).chunk=NULL;
	(*marker).chunknum=$tail->chunknum+1;
	(*marker).pos=0;
	/* search vertically down the page */
	prev=$pages[$currentPage].start;
	for (c=$pages[$currentPage].start; c!=NULL; prev=c, c = c->next) {
		if (c->type==CT_TRAVERSER)
			if (c->pagenum!=$currentPage)
				break;

		if (y <= c->y+c->depth)
			break;
		}

	if (c==NULL)
		return;

	/* is the mouse pointer between lines? */
	if (y <= c->y - c->height) {
		(*marker).chunk=prev;
		(*marker).chunknum=prev->chunknum;
		if (prev->text != NULL)
			(*marker).pos=strlen(prev->text)+1;
		else
			(*marker).pos=0;
		return;
		}

	/* search horizontally across the page */
	cury =c->y;
	for ( ; (c!=NULL) && (cury==c->y); prev=c, c = c->next) {
		if (x <= c->x+c->width)
			break;
		}
	if (c==NULL)
		return;

	/* at the end of the line? */
	if (cury!=c->y) {
		(*marker).chunk=prev;
		(*marker).chunknum=prev->chunknum;
		if (prev->text != NULL)
			(*marker).pos=strlen(prev->text)+1;
		else
			(*marker).pos=0;
		return;
		}

	/* between chunks? */
	if (x < c->x) {
		(*marker).chunk=c;
		(*marker).chunknum=c->chunknum;
		return;
		}

	if (c->type==CT_LINEBREAK) {
		if (prev->text != NULL) {
			(*marker).chunk=prev;
			(*marker).chunknum=prev->chunknum;
			(*marker).pos=strlen(prev->text)+1;
			}
		else {
			(*marker).chunk=c;
			(*marker).chunknum=c->chunknum;
			}
		return;
		}

	/* treat a dottab/bullet/traverser as a single chunk */
	if ((c->type==CT_DOTTAB) || (c->type==CT_BULLET) || (c->type==CT_TRAVERSER)) {
		(*marker).chunknum = c->chunknum + 1;
		c=c->next;  /* assumes we never get two dottabs/bullets in a row */
		(*marker).chunk=c;
		return;
		}

	switch(c->font)     {
		case 2: f=$font2; break;
		case 3: f=$font3; break;
		case 4: f=$font4; break;
		case 5: f=$font5; break;
		default: f=$font1; break;
		}
	slen=strlen(c->text);
	if (newline(c->text))
		slen--;
	
	for (i=1; i<slen; i++) {
		xpos=c->x + XTextWidth(f,c->text,i);
		if (xpos>=x)
			break;
		}

	(*marker).chunk=c;
	(*marker).chunknum=c->chunknum;
	(*marker).pos=i-1;
	}



@ Returns the minimum (earlier) of two positions in the text.

@proc SelectionMarker minimum_marker(SelectionMarker a, SelectionMarker b) {
	if (a.chunknum==b.chunknum) {
		return ((a.pos < b.pos) ? a : b);
		}
	else if (a.chunknum < b.chunknum)
		return a;
	return b;
	}



@ Returns the maximum (later) of two positions in the text.

@proc SelectionMarker maximum_marker(SelectionMarker a, SelectionMarker b) {
	if (a.chunknum==b.chunknum) {
		return ((a.pos > b.pos) ? a : b);
		}
	else if (a.chunknum > b.chunknum)
		return a;
	return b;
	}



@ Returns |True| if the two text positions are coincident.

@proc Boolean equal_marker(SelectionMarker a, SelectionMarker b) {
	if (a.chunknum!=b.chunknum) 
		return False;
	return (a.pos==b.pos);
	}



@ Returns |True| if position |a| is smaller (earlier) than position |b|.

@proc Boolean smaller(SelectionMarker a, SelectionMarker b) {
	if (a.chunknum<b.chunknum) 
		return True;
	if (a.chunknum>b.chunknum) 
		return False;
	return (a.pos<=b.pos);
	}


@ Timer callback procedure - scroll display by one line and recompute
selection endpoint.

@proc scroll_display_up(XtPointer client_data, XtIntervalId *timer) {
    Widget $ = (Widget) client_data;
	Window rwin, chwin;
	int rx, ry, wx, wy;
	unsigned int mask;

	/* scroll the display up one line */
	XtCallActionProc(XtParent($), "lineDown", CurrentTime, NULL, 0);

	/* recompute the selection end point */
	XQueryPointer(XtDisplay($), XtWindow($), &rwin, &chwin, &rx, &ry,
		&wx, &wy, &mask);
	find_position($, wx, wy, &($selectEnd));

	/* set the timer running again */
	$scrollTimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
		$scrollInterval, scroll_display_up, $);
        }

@ Timer callback procedure - scroll display by one line and recompute
selection endpoint.

@proc scroll_display_down(XtPointer client_data, XtIntervalId *timer) {
    Widget $ = (Widget) client_data;
	Window rwin, chwin;
	int rx, ry, wx, wy;
	unsigned int mask;

	/* scroll the display up one line */
	XtCallActionProc(XtParent($), "lineUp", CurrentTime, NULL, 0);

	/* recompute the selection end point */
	XQueryPointer(XtDisplay($), XtWindow($), &rwin, &chwin, &rx, &ry,
		&wx, &wy, &mask);
	find_position($, wx, wy, &($selectEnd));

	/* set the timer running again */
	$scrollTimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
		$scrollInterval, scroll_display_down, $);
        }


@ACTIONS

@ the |chaselink| action invokes the |activate| callback if the mouse pointer
is over a hyperlink and text selection is not in progress.

@proc chaselink {
	int mx, my;
	TextChunk *c;
	static LinkRecord lr;
	
	if ($selecting)
		return;

	mx=event->xbutton.x;
	my=event->xbutton.y;


	for (c=$pages[$currentPage].start; c!=NULL; c = c->next) {
		if ((c->type != CT_LINK) && (c->type != CT_TRAVERSER))
			continue;
		if (c->type == CT_TRAVERSER)
			if (c->pagenum != $currentPage)
				break;

		if ((mx >= c->x) && (mx <= c->x + c->width) && 
			(my <= c->y + c->depth) && (my >= c->y - c->height)) {
			lr.text = c->text;
			lr.link = c->link;
			if (c->type == CT_TRAVERSER) {
				if (mx > c->x + trav_width/2)
					LinkTextNextPage($, True);
				else
					LinkTextPrevPage($, True);
				}
			else
				XtCallCallbackList($, $activate, &lr);
			return;
			}
		}
	}

@ The |grabpage| action causes the mouse cursor to change into a hand shape.
Subsequent mouse movements will allow the screen to be scrolled via the
|movepage| action.

@proc grabpage {
    $grabx=event->xbutton.x_root;
    $graby=event->xbutton.y_root;
	LinkViewGetCoordinates(XtParent($), &($viewx), &($viewy));
    XDefineCursor(XtDisplay($), XtWindow($), $grabCursor);
	}

@ The |movepage| action causes the display to be scrolled within the
parent viewport in response to mouse movements.

@proc movepage {
	int mx, my;

    mx=event->xbutton.x_root;
    my=event->xbutton.y_root;
	LinkViewSetCoordinates(XtParent($), $viewx-mx+$grabx, $viewy-my+$graby);
	}

@ The |releasepage| action performs various cleanup operations needed
after the display has been scrolled via the |movepage| action.


@proc releasepage {
    XDefineCursor(XtDisplay($), XtWindow($), $normalCursor);
	}

@ The |cursortype| action sets the cursor to |normalCursor| or |linkCursor|
depending on whether the mouse pointer is over a hyperlink.

@proc cursortype {
	int mx, my;
	TextChunk *c;
	
	if ($busy) {
        XDefineCursor(XtDisplay($), XtWindow($), $busyCursor);
		return;
		}

	mx=event->xbutton.x;
	my=event->xbutton.y;

	for (c=$pages[$currentPage].start; c!=NULL; c = c->next) {
		if ((c->type != CT_LINK) && (c->type != CT_TRAVERSER))
			continue;

		/* check for end of page */
		if (c->type == CT_TRAVERSER)
			if (c->pagenum != $currentPage)
				break;

		if ((mx >= c->x) && (mx <= c->x + c->width) && 
			(my <= c->y + c->depth) && (my >= c->y - c->height)) {
	        XDefineCursor(XtDisplay($), XtWindow($), $linkCursor);
			return;
			}
		}
    XDefineCursor(XtDisplay($), XtWindow($), $normalCursor);
	}



@ Start text selection. (Usually bound to mouse button 1.)

@proc start_selection {
	if ($head==NULL)
		return;

	/* don't start selecting until the mouse has moved a certain number
	   of pixels */
	$selecting=False;
	clear_selection($);


	/* remember where the mouse is */
    $selectx_root=event->xbutton.x_root;
    $selecty_root=event->xbutton.y_root;
    $selectx=event->xbutton.x;
    $selecty=event->xbutton.y;

	find_position($, $selectx, $selecty, &($selectStart));
	}



@ Drag out the text selection. (Usually bound to motion while mouse button 1
is held down.

@proc move_selection {
	int mx_root, my_root;
	int mx, my;
	SelectionMarker top, bottom;
	XExposeEvent ev;
	Boolean selection_started=False;
	Dimension lvh;
	int lvx, lvy;

	if ($head==NULL)
		return;

    mx_root=event->xbutton.x_root;
    my_root=event->xbutton.y_root;
	if (! $selecting)
		/* has the mouse moved far enough to start selection? */
		if ((abs(mx_root-$selectx_root) >= $selectThreshold) ||
			(abs(my_root-$selecty_root) >= $selectThreshold))	{
			$selected=True;
			$selecting=True;
			selection_started=True;
		    XDefineCursor(XtDisplay($), XtWindow($), $normalCursor);

			/* find y position in root window of top and bottom border */
			LinkViewGetCoordinates(XtParent($), &lvx, &lvy);
			XtVaGetValues(XtParent($), XtNheight, &lvh, NULL);
			$topBorder=my_root - event->xbutton.y + lvy;
			$bottomBorder=$topBorder+lvh;
			}
	if (! $selecting)
		return;

    mx=event->xbutton.x;
    my=event->xbutton.y;
	find_position($, mx, my, &($selectEnd));


	/* scroll the display if we need to */

	if (my_root<$topBorder) {
		if (! $scrollTimer) {
			$scrollTimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
				$scrollInterval, scroll_display_down, $);
			}
		}
	else if (my_root>$bottomBorder) {
		if (! $scrollTimer) {
			$scrollTimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
				$scrollInterval, scroll_display_up, $);
			}
		}
	else if ($scrollTimer) {
		XtRemoveTimeOut($scrollTimer);
		$scrollTimer=0;
		}


	/* if the selection end has moved (or just begun) we need to update the display */

	if (!equal_marker($selectEnd, $lastend) || selection_started) {
		/* printf("selectEnd=%d,%d\n", $selectEnd.chunknum, $selectEnd.pos);*/
		if (selection_started) {
			top=minimum_marker($selectStart, $selectEnd);
			bottom=maximum_marker($selectStart, $selectEnd);
			}
		else {
			top=minimum_marker($selectEnd, $lastend);
			bottom=maximum_marker($selectEnd, $lastend);
			}

		/* construct an exposure bounding box */
		ev.x=0;
		ev.width=$width;
		ev.y=top.chunk->y - top.chunk->height-5;
		if (bottom.chunk==NULL)
			ev.height=$height - ev.y;
		else
			ev.height=bottom.chunk->y + bottom.chunk->depth - ev.y;
		expose($, (XEvent*) &ev, NULL);

		/* remember where we last updated the screen */
		$lastend=$selectEnd;
		}
	}



@ Extend the text selection. (Usually bound to mouse button 3.)

@proc extend_selection {
	int mx, my;
	SelectionMarker newend, top, bottom;
	XExposeEvent ev;

	if (($selectStart.chunknum<=0) || ($head==NULL))
		return;

    mx=event->xbutton.x;
    my=event->xbutton.y;
	find_position($, mx, my, &newend);

	if (!$selected) {
		$selected=True;
		if (XtOwnSelection($, XA_PRIMARY, event->xbutton.time, 
			convert_proc, lose_ownership_proc, convert_done_proc) == False) {
			XtWarning("LinkText: failed to become selection owner\n");
			return;
			}
		$selectEnd=newend;
		top=minimum_marker($selectStart, $selectEnd);
		bottom=maximum_marker($selectStart, $selectEnd);
		}
	else if (smaller(newend, $selectStart) && smaller($selectStart, $selectEnd)) {
		top=newend;
		bottom=$selectStart;
		$selectStart=newend;
		}
	else if (smaller($selectStart, newend) && smaller(newend, $selectEnd)) {
		top=newend;
		bottom=$selectEnd;
		$selectEnd=newend;
		}
	else if (smaller($selectStart, $selectEnd) && smaller($selectEnd, newend)) {
		top=$selectEnd;
		bottom=newend;
		$selectEnd=newend;
		}
	else if (smaller(newend, $selectEnd) && smaller($selectEnd, $selectStart)) {
		top=newend;
		bottom=$selectEnd;
		$selectEnd=newend;
		}
	else if (smaller($selectEnd, newend) && smaller(newend, $selectStart)) {
		top=$selectEnd;
		bottom=newend;
		$selectEnd=newend;
		}
	else {
		top=$selectStart;
		bottom=newend;
		$selectStart=newend;
		}

	if (top.chunk==NULL) {
		XtError("extend_selection error - contact author");
		return;
		}

	/* construct an exposure bounding box */
	ev.x=0;
	ev.width=$width;
	ev.y=top.chunk->y - top.chunk->height;
	if (bottom.chunk==NULL)
		ev.height=$height - ev.y;
	else
		ev.height=bottom.chunk->y + bottom.chunk->depth - ev.y;

	expose($, (XEvent*) &ev, NULL);
	}

@ Finish making a text selection with the mouse.

@proc end_selection {
	if (!$selecting)
		return;

	if ($scrollTimer) {
		XtRemoveTimeOut($scrollTimer);
		$scrollTimer=0;
		}

	$selecting=False;
	if (XtOwnSelection($, XA_PRIMARY, event->xbutton.time, 
		convert_proc, lose_ownership_proc, convert_done_proc) == False) {
		XtWarning("LinkText: failed to become selection owner\n");
		return;
		}
	$selected=True;
	}



@UTILITIES

@ Clears the text selection if it was active.

@proc clear_selection($) {
	XExposeEvent ev;
	SelectionMarker top, bottom;

	if (!$selected) 
		return;

	top=minimum_marker($selectStart, $selectEnd);
	bottom=maximum_marker($selectStart, $selectEnd);

	ev.x=0;
	ev.width=$width;
	ev.y=top.chunk->y - top.chunk->height;
	if (bottom.chunk==NULL)
		ev.height=$height - ev.y;
	else
		ev.height=bottom.chunk->y + bottom.chunk->depth - ev.y;

	/* Note: if I wanted to be a real smarty-pants I could trim the exposure
 	   region according to what's visible in the parent viewport - in practice
	   this works well enough. */

	$selected=False;
	expose($, (XEvent*) &ev, NULL);
	}



@ Called if another widget obtains the ownership of the PRIMARY atom.

@proc lose_ownership_proc($, Atom *selection) {
	clear_selection($);
	$selectStart.chunknum=0;
	}

@ Called when a text selection transfer has been completed.

@proc convert_done_proc($, Atom *selection,  Atom *target) {
	/* we could free the asciitext buffer if we wanted to */
	}



@ The procedure to convert the current text selection into plain ASCII text.

@proc Boolean convert_proc($, Atom *selection, Atom *target, Atom *type, 
        XtPointer *value, unsigned long *length, int *format) {
#ifdef SVR4                                     /* BB, 27 Mar '96 */
#define bcopy(b1, b2, length) memcpy(b2, b1, length)
#endif
    Display* d = XtDisplay($);
    XSelectionRequestEvent* req = XtGetSelectionRequest($, *selection, NULL);

    if (*target == XA_TARGETS(d)) {
        Atom* targetP;
        Atom* std_targets;
        unsigned long std_length;
        XmuConvertStandardSelection($, req->time, selection, target,
                                    type, (XPointer*)&std_targets,
                                    &std_length, format);
        *value = XtMalloc(sizeof(Atom) * (std_length + 1));
        targetP = *(Atom**) value;
        *targetP++ = XA_STRING;
        *length = std_length + (targetP - (*(Atom **) value));
        bcopy((char*)std_targets, (char*)targetP, sizeof(Atom)*std_length);
        XtFree((char*)std_targets);
        *type = XA_ATOM;
        *format = 32;   /* ORA code example bombs out on 64 bit platforms */
        return True;
    }
    if (*target == XA_STRING) {
        *type = XA_STRING;
        get_text($, False);
        *value = $asciitext->buf;
		if (*value==NULL)
			return False;
        *length = sizeof(char) * (strlen(*value));
        *format = 8;
        return True;
    }
    if (XmuConvertStandardSelection($, req->time, selection, target,
                                    type, (XPointer*) value, length, format))
        return True;

    return False;
}



@ Adds the given text to the ASCII buffer, wrapping it so that no line
exceeds 80 characters.

@proc SmartAddToBuffer($, int *x, char *str, int font, Boolean inbullet, 
		Boolean need_indent) {
	char *cp, *start;

	cp=str;
	start=str;
	while (*cp != '\0') {
		if (*cp==' ') {
			if (*x + cp-start+1> 78) {
				AddToBuffer($asciitext, "\n");
				*x=0;
				if (inbullet) {
					AddToBuffer($asciitext, "   ");
					*x=3;
					}
				}
			/* indent 'normal' text - makes headings stand out */
			if ((*x==0) && (font<=2) && need_indent) {
				AddToBuffer($asciitext, "  ");
				*x += 2;
				}
			AddToBufferN($asciitext, start, cp-start+1);
			*x += cp-start+1;
			start=cp+1;
			}
		cp++;
		}

	/* tack on the final piece */
	if (*start != '\0') {
		if (*x + cp-start> 78) {
			AddToBuffer($asciitext, "\n");
			*x=0;
			if (inbullet) {
				AddToBuffer($asciitext, "   ");
				*x=3;
				}
			}
		/* indent 'normal' text - makes headings stand out */
		if ((*x==0) && (font<=2) && need_indent) {
			AddToBuffer($asciitext, "  ");
			*x += 2;
			}
		AddToBufferN($asciitext, start, cp-start);
		*x += cp-start;
		}
	}

@ This collects the selected text into the |asciitext| buffer.

@proc get_text($, Boolean getall) {
	TextChunk *c, *start;
	SelectionMarker top, bottom;
	int x, tab;
	Boolean inbullet=False;
	Boolean need_indent;

	if ($asciitext==NULL)
		$asciitext=CreateBuffer("LinkTxt", 512, 512, 262144);

	ClearBuffer($asciitext);

	if ($head==NULL)
		return;

	if (getall) {
		top.chunk=$head;
		top.chunknum=1;
		top.pos=0;
		bottom.chunk=$tail;
		bottom.chunknum=$tail->chunknum;
		if (($tail->type == CT_LINK) || ($tail->type == CT_TEXT))
			bottom.pos=strlen($tail->text);
		else
			bottom.pos=0;
		}
	else {
		if (!$selected)
			return;
		if (equal_marker($selectStart, $selectEnd))
			return;
		top=minimum_marker($selectStart, $selectEnd);
		bottom=maximum_marker($selectStart, $selectEnd);
		}

	start=top.chunk;
	if (start==NULL)
		return;

	if (top.chunknum==bottom.chunknum) {
		if (top.chunk->type==CT_DOTTAB)
			AddToBuffer($asciitext, " ..... ");
		else
			AddToBufferN($asciitext, top.chunk->text+top.pos, 
			bottom.pos-top.pos);
		return;
		}

	/* see if there are any headings to be converted */
	need_indent=False;
	if ($indentAscii)
		for (c=start; c!=NULL; c = c->next) {
			if (c->chunknum > bottom.chunknum)
				break;
			if ((c->type==CT_TEXT) && (c->font > 2)) {
				need_indent=True;
				break;
				}
			}

	x=0;
	for (c=start; c!=NULL; c = c->next) {
		if (c->chunknum > bottom.chunknum)
			break;
		switch (c->type) {
			case CT_LINK:
			case CT_TEXT:
				/* indent 'normal' text - makes headings stand out */
				if ((x==0) && (c->font<=2) && need_indent) {
					AddToBuffer($asciitext, "  ");
					x += 2;
					}
				if (c->chunknum == top.chunknum) 
					SmartAddToBuffer($, &x, c->text + top.pos, c->font, inbullet, need_indent);
				else if (c->chunknum == bottom.chunknum) {
					if (bottom.pos>0) {
						AddToBufferN($asciitext, c->text, bottom.pos);
						}
					}
				else 
					SmartAddToBuffer($, &x, c->text, c->font, inbullet, need_indent);
				if (newline(c->text))
					x=0;
				break;
			case CT_BULLET:
				AddToBuffer($asciitext, " * ");
				x=3;
				inbullet=True;
				break;
			case CT_LINEBREAK:
				AddToBuffer($asciitext, "\n");
				inbullet=False;
				x=0;
				break;
			case CT_DOTTAB:
				tab=c->indent/$spaceWidth;
				if (x>tab-5) {
					AddToBuffer($asciitext, "\n");
					x=0;
					while (x<tab/2) {
						AddToBuffer($asciitext, " ");
						x++;
						}
					}
				AddToBuffer($asciitext, " ");
				x++;
				while (x<tab-1) {
					AddToBuffer($asciitext, ".");
					x++;
					}
				AddToBuffer($asciitext, " ");
				x=tab;
				break;
			case CT_TRAVERSER:
				/* ignore page traversers */
				break;
			default:
				XtWarning("get_selected_text: unhandled case");
			}
		}
	}



@EXPORTS

@ The |HypertextToText| procedure returns a pointer to a memory block 
containing an ASCII text version of the hypertext
currently displayed by the widget. 

@proc char* HypertextToText($) {
	get_text($, True);
	return $asciitext->buf;
	}


@ The |SetBusy| procedure takes one boolean argument. If it is true, then the widget
will display the |busyCursor|, otherwise it will display the |normalCursor|
or |linkCursor| as appropriate.

@proc SetBusy($, Boolean v) {
	$busy=v;

	if ($busy)
	    XDefineCursor(XtDisplay($), XtWindow($), $busyCursor);
	else
	    XDefineCursor(XtDisplay($), XtWindow($), $normalCursor);
	}



@ The |GetBaselineskip| procedure returns number of pixels between lines of text.  
The procedure could be really tricky, but at the moment it simply
assumes that |font1| is the dominant text and returns the baselineskip 
for that font.

@proc Dimension GetBaselineskip($) {
	return ($font1->ascent + $font1->descent);
	}



@ When a display consists of more than one page, display the first page.

@proc LinkTextFirstPage($, Boolean reset_to_top) {
	LinkTextSetPage($, 0, reset_to_top);
	}


@ When a display consists of more than one page, display the last page.

@proc LinkTextLastPage($, Boolean reset_to_top) {
	LinkTextSetPage($, $numPages-1, reset_to_top);
	}

@ When a display consists of more than one page, display the next page
if it exists.

@proc LinkTextNextPage($, Boolean reset_to_top) {
	LinkTextSetPage($, $currentPage + 1, reset_to_top);
	}



@ When a display consists of more than one page, display the previous page
if it exists.

@proc LinkTextPrevPage($, Boolean reset_to_top) {
	LinkTextSetPage($, $currentPage - 1, reset_to_top);
	}



@ When a display consists of more than one page, display the specified page
if it exists.

@proc LinkTextSetPage($, int p, Boolean reset_to_top) {
	Dimension parent_width, parent_height, width;
	static PageRecord pr;

	if ((p<0) || (p>=$numPages))
		return;
	$currentPage = p;

	/* resize the widget - but don't make it narrower than the parent */
	width = $pages[$currentPage].width;
	XtVaGetValues(XtParent($), XtNwidth, &parent_width, 
		XtNheight, &parent_height, NULL);
	if (width < parent_width)
		width=parent_width;
	XtResizeWidget($, width, $pages[$currentPage].height, 0);
	LinkViewSetCoordinates(XtParent($), 0, 0);
	if (!reset_to_top) 
		XtCallActionProc(XtParent($), "pageBottom", NULL, NULL, 0);

	XClearArea(XtDisplay($), XtWindow($), 0, -$y, $width, parent_height, True);

 	pr.current=$currentPage;
	pr.total=$numPages;
	XtCallCallbackList($, $newpage, &pr);
	}


@IMPORTS

@incl <assert.h>
@incl <ctype.h>
@incl <stdio.h>
@incl <string.h>
@incl "buffer.h"
@incl "LinkView.h"
@incl <X11/Xatom.h>
@incl <X11/Xmu/Xmu.h>
@incl "hand.bm"
@incl "handmask.bm"
@incl "trav.xbm"
@incl "stip4.bm"
