/* This a modified version of the Athena Viewport widget for use with the
   LinkText hyperlink widget. It was modified by Lachlan Wetherall in 
   July 1996.  

Changes from the original code:

   05-Jul-96  * Added keyboard actions.
              * Added baselineskip resource.
              * Resize sets the width of the child widget before attempting
                to compute the layout. This allows the child widget to
                perform auto word wrapping.
              * Removed XawViewportSetLocation() and 
                XawViewportSetCoordinates() and added LinkViewReload() and
                LinkViewGetCoordinates().
	11-Oct-96 * Added childResizable resource.

   The copyright notices from the original widget code appears below. 

*/

/***********************************************************

Copyright (c) 1987, 1988, 1994  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.


Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/XawInit.h>
#include <X11/Xmu/Misc.h>
#include <X11/Xaw/Scrollbar.h>
#include "LinkViewP.h"
#include "LinkText.h"

static void ScrollUpDownProc(), ThumbProc();
static Boolean GetGeometry();

static void ComputeWithForceBars();

#define offset(field) XtOffsetOf(LinkViewRec, linkview.field)
static XtResource resources[] = {
    {XtNforceBars, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(forcebars), XtRImmediate, (XtPointer)False},
    {XtNallowHoriz, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(allowhoriz), XtRImmediate, (XtPointer)False},
    {XtNallowVert, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(allowvert), XtRImmediate, (XtPointer)False},
    {XtNuseBottom, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(usebottom), XtRImmediate, (XtPointer)False},
    {XtNuseRight, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(useright), XtRImmediate, (XtPointer)False},
    {XtNreportCallback, XtCReportCallback, XtRCallback, sizeof(XtPointer),
	 offset(report_callbacks), XtRImmediate, (XtPointer) NULL},
    {XtNbaselineskip, XtCBaselineskip, XtRDimension, sizeof(Dimension),
	 offset(baselineskip), XtRImmediate, (XtPointer) 20},
    {XtNchildResizable, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(child_resizable), XtRImmediate, (XtPointer)False},
};
#undef offset

static void Initialize(), ConstraintInitialize(),
    Realize(), Resize(), ChangeManaged();
static Boolean SetValues(), Layout();
static XtGeometryResult GeometryManager(), PreferredGeometry();

static char defaultTranslations[] =
	"	<Key>Left:       pageLeft()\n\
		<Key>Right:      pageRight()\n\
		Shift<Key>Prior: pagePrev()\n\
		Shift<Key>Next:  pageNext()\n\
		Ctrl<Key>Prior:  pageTop()\n\
		Ctrl<Key>Next:   pageBottom()\n\
		Ctrl<Key>Home:	 pageFirst()\n\
		Ctrl<Key>End:	 pageLast()\n\
		<Key>space:      pageDown()\n\
		<Key>-:          pageUp()\n\
		<Key>Next:       pageDown()\n\
		<Key>Prior:      pageUp()\n\
		<Key>Up:         lineUp()\n\
		<Key>Down:       lineDown()";

static void lineDown();
static void lineUp();
static void pagePrev();
static void pageNext();
static void pageDown();
static void pageUp();
static void pageFirst();
static void pageLast();
static void pageBottom();
static void pageTop();
static void pageLeft();
static void pageRight();

static XtActionsRec actionsList[] = {
		{"lineDown", lineDown},
		{"lineUp", lineUp},
		{"pagePrev", pagePrev},
		{"pageNext", pageNext},
		{"pageDown", pageDown},
		{"pageUp", pageUp},
		{"pageBottom", pageBottom},
		{"pageTop", pageTop},
		{"pageLeft", pageLeft},
		{"pageRight", pageRight},
		{"pageFirst", pageFirst},
		{"pageLast", pageLast},
		};

#define superclass	(&formClassRec)
LinkViewClassRec linkViewClassRec = {
  { /* core_class fields */
    /* superclass	  */	(WidgetClass) superclass,
    /* class_name	  */	"LinkView",
    /* widget_size	  */	sizeof(LinkViewRec),
    /* class_initialize	  */	XawInitializeWidgetSet,
    /* class_part_init    */    NULL,
    /* class_inited	  */	FALSE,
    /* initialize	  */	Initialize,
    /* initialize_hook    */    NULL,
    /* realize		  */	Realize,
    /* actions		  */	actionsList,
    /* num_actions	  */	XtNumber(actionsList),
    /* resources	  */	resources,
    /* num_resources	  */	XtNumber(resources),
    /* xrm_class	  */	NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/    TRUE,
    /* visible_interest	  */	FALSE,
    /* destroy		  */	NULL,
    /* resize		  */	Resize,
    /* expose		  */	XtInheritExpose,
    /* set_values	  */	SetValues,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus	  */	NULL,
    /* version            */    XtVersion,
    /* callback_private	  */	NULL,
    /* tm_table    	  */	defaultTranslations,
    /* query_geometry     */    PreferredGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },
  { /* composite_class fields */
    /* geometry_manager	  */	GeometryManager,
    /* change_managed	  */	ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },
  { /* constraint_class fields */
    /* subresourses	  */	NULL,
    /* subresource_count  */	0,
    /* constraint_size	  */	sizeof(LinkViewConstraintsRec),
    /* initialize	  */	ConstraintInitialize,
    /* destroy		  */	NULL,
    /* set_values	  */	NULL,
    /* extension          */	NULL
  },
  { /* form_class fields */
    /* layout		  */	Layout
  },
  { /* linkview_class fields */
    /* empty		  */	0
  }
};


WidgetClass linkViewWidgetClass = (WidgetClass)&linkViewClassRec;

static Widget CreateScrollbar(w, horizontal)
    LinkViewWidget w;
    Boolean horizontal;
{
    Widget clip = w->linkview.clip;
    LinkViewConstraints constraints =
	(LinkViewConstraints)clip->core.constraints;
    static Arg barArgs[] = {
	{XtNorientation,       (XtArgVal) 0},
	{XtNlength,            (XtArgVal) 0},
	{XtNleft,              (XtArgVal) 0},
	{XtNright,             (XtArgVal) 0},
	{XtNtop,               (XtArgVal) 0},
	{XtNbottom,            (XtArgVal) 0},
	{XtNmappedWhenManaged, (XtArgVal) False},
    };
    Widget bar;

    XtSetArg(barArgs[0], XtNorientation,
	      horizontal ? XtorientHorizontal : XtorientVertical );
    XtSetArg(barArgs[1], XtNlength,
	     horizontal ? clip->core.width : clip->core.height);
    XtSetArg(barArgs[2], XtNleft,
	     (!horizontal && w->linkview.useright) ? XtChainRight : XtChainLeft);
    XtSetArg(barArgs[3], XtNright,
	     (!horizontal && !w->linkview.useright) ? XtChainLeft : XtChainRight);
    XtSetArg(barArgs[4], XtNtop,
	     (horizontal && w->linkview.usebottom) ? XtChainBottom: XtChainTop);
    XtSetArg(barArgs[5], XtNbottom,
	     (horizontal && !w->linkview.usebottom) ? XtChainTop: XtChainBottom);

    bar = XtCreateWidget((horizontal ? "horizontal" : "vertical"),
			  scrollbarWidgetClass, (Widget)w,
			  barArgs, XtNumber(barArgs) );
    XtAddCallback( bar, XtNscrollProc, ScrollUpDownProc, (XtPointer)w );
    XtAddCallback( bar, XtNjumpProc, ThumbProc, (XtPointer)w );

    if (horizontal) {
	w->linkview.horiz_bar = bar;
	constraints->form.vert_base = bar;
    }
    else {
	w->linkview.vert_bar = bar;
	constraints->form.horiz_base = bar;
    }

    XtManageChild( bar );

    return bar;
}

/* ARGSUSED */
static void Initialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    LinkViewWidget w = (LinkViewWidget)new;
    static Arg clip_args[8];
    Cardinal arg_cnt;
    Widget h_bar, v_bar;
    Dimension clip_height, clip_width;

    w->form.default_spacing = 0;  /* Reset the default spacing to 0 pixels. */


/* 
 * Initialize all widget pointers to NULL.
 */

    w->linkview.child = (Widget) NULL;
    w->linkview.horiz_bar = w->linkview.vert_bar = (Widget)NULL;

/* 
 * Create Clip Widget.
 */

    arg_cnt = 0;
    XtSetArg(clip_args[arg_cnt], XtNbackgroundPixmap, None); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNborderWidth, 0); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNleft, XtChainLeft); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNright, XtChainRight); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNtop, XtChainTop); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNbottom, XtChainBottom); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNwidth, w->core.width); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNheight, w->core.height); arg_cnt++;

    w->linkview.clip = XtCreateManagedWidget("clip", widgetClass, new,
					     clip_args, arg_cnt);

    if (!w->linkview.forcebars) 
        return;		 /* If we are not forcing the bars then we are done. */

    if (w->linkview.allowhoriz) 
      (void) CreateScrollbar(w, True);
    if (w->linkview.allowvert) 
      (void) CreateScrollbar(w, False);

    h_bar = w->linkview.horiz_bar;
    v_bar = w->linkview.vert_bar;

/*
 * Set the clip widget to the correct height.
 */

    clip_width = w->core.width;
    clip_height = w->core.height;

    if ( (h_bar != NULL) &&
	 ((int)w->core.width >
	  (int)(h_bar->core.width + h_bar->core.border_width)) )
        clip_width -= h_bar->core.width + h_bar->core.border_width;
    
    if ( (v_bar != NULL) &&
	 ((int)w->core.height >
	  (int)(v_bar->core.height + v_bar->core.border_width)) )
        clip_height -= v_bar->core.height + v_bar->core.border_width;

    arg_cnt = 0;
    XtSetArg(clip_args[arg_cnt], XtNwidth, clip_width); arg_cnt++;
    XtSetArg(clip_args[arg_cnt], XtNheight, clip_height); arg_cnt++;
    XtSetValues(w->linkview.clip, clip_args, arg_cnt);
}

/* ARGSUSED */
static void ConstraintInitialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    ((LinkViewConstraints)new->core.constraints)->linkview.reparented = False;
}

static void Realize(widget, value_mask, attributes)
    Widget widget;
    XtValueMask *value_mask;
    XSetWindowAttributes *attributes;
{
    LinkViewWidget w = (LinkViewWidget)widget;
    Widget child = w->linkview.child;
    Widget clip = w->linkview.clip;

    *value_mask |= CWBitGravity;
    attributes->bit_gravity = NorthWestGravity;
    (*superclass->core_class.realize)(widget, value_mask, attributes);

/* printf("LinkView: Realize\n");*/
    (*w->core.widget_class->core_class.resize)(widget);	/* turn on bars */

    if (child != (Widget)NULL) {
	XtMoveWidget( child, (Position)0, (Position)0 );
	XtRealizeWidget( clip );
	XtRealizeWidget( child );
	XReparentWindow( XtDisplay(w), XtWindow(child), XtWindow(clip),
			 (Position)0, (Position)0 );
	XtMapWidget( child );
    }
}

/* ARGSUSED */
static Boolean SetValues(current, request, new, args, num_args)
    Widget current, request, new;
    ArgList args;
    Cardinal *num_args;
{
    LinkViewWidget w = (LinkViewWidget)new;
    LinkViewWidget cw = (LinkViewWidget)current;

    if ( (w->linkview.forcebars != cw->linkview.forcebars) ||
	 (w->linkview.allowvert != cw->linkview.allowvert) ||
	 (w->linkview.allowhoriz != cw->linkview.allowhoriz) ||
	 (w->linkview.useright != cw->linkview.useright) ||
	 (w->linkview.usebottom != cw->linkview.usebottom) ) 
    {
/* printf("LinkView: SetValues:\n");*/
	(*w->core.widget_class->core_class.resize)(new); /* Recompute layout.*/
    }

    return False;
}


static void ChangeManaged(widget)
    Widget widget;
{
    LinkViewWidget w = (LinkViewWidget)widget;
    int num_children = w->composite.num_children;
    Widget child, *childP;
    int i;

    child = (Widget)NULL;
    for (childP=w->composite.children, i=0; i < num_children; childP++, i++) {
	if (XtIsManaged(*childP)
	    && *childP != w->linkview.clip
	    && *childP != w->linkview.horiz_bar
	    && *childP != w->linkview.vert_bar)
	{
	    child = *childP;
	    break;
	}
    }

    if (child != w->linkview.child) {
	w->linkview.child = child;
	if (child != (Widget)NULL) {
	    XtResizeWidget( child, child->core.width,
			    child->core.height, (Dimension)0 );
	    if (XtIsRealized(widget)) {
		LinkViewConstraints constraints =
		    (LinkViewConstraints)child->core.constraints;
		if (!XtIsRealized(child)) {
		    Window window = XtWindow(w);
		    XtMoveWidget( child, (Position)0, (Position)0 );
#ifdef notdef
		    /* this is dirty, but it saves the following code: */
		    XtRealizeWidget( child );
		    XReparentWindow( XtDisplay(w), XtWindow(child),
				     XtWindow(w->linkview.clip),
				     (Position)0, (Position)0 );
		    if (child->core.mapped_when_managed)
			XtMapWidget( child );
#else 
		    w->core.window = XtWindow(w->linkview.clip);
		    XtRealizeWidget( child );
		    w->core.window = window;
#endif /* notdef */
		    constraints->linkview.reparented = True;
		}
		else if (!constraints->linkview.reparented) {
		    XReparentWindow( XtDisplay(w), XtWindow(child),
				     XtWindow(w->linkview.clip),
				     (Position)0, (Position)0 );
		    constraints->linkview.reparented = True;
		    if (child->core.mapped_when_managed)
			XtMapWidget( child );
		}
	    }
	    GetGeometry( widget, child->core.width, child->core.height );
	    (*((LinkViewWidgetClass)w->core.widget_class)->form_class.layout)
		( (FormWidget)w, w->core.width, w->core.height );
	    /* %%% do we need to hide this child from Form?  */
	}
    }

#ifdef notdef
    (*superclass->composite_class.change_managed)( widget );
#endif
}


static void SetBar(w, top, length, total)
    Widget w;
    Position top;
    Dimension length, total;
{
    XawScrollbarSetThumb(w, (float)top/(float)total,
			 (float)length/(float)total);
}

static void RedrawThumbs(w)
  LinkViewWidget w;
{
    Widget child = w->linkview.child;
    Widget clip = w->linkview.clip;

    if (w->linkview.horiz_bar != (Widget)NULL)
	SetBar( w->linkview.horiz_bar, -(child->core.x),
	        clip->core.width, child->core.width );

    if (w->linkview.vert_bar != (Widget)NULL)
	SetBar( w->linkview.vert_bar, -(child->core.y),
	        clip->core.height, child->core.height );
}



static void SendReport (w, changed)
    LinkViewWidget w;
    unsigned int changed;
{
    XawPannerReport rep;

    if (w->linkview.report_callbacks) {
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	rep.changed = changed;
	rep.slider_x = -child->core.x;	/* child is canvas */
	rep.slider_y = -child->core.y;	/* clip is slider */
	rep.slider_width = clip->core.width;
	rep.slider_height = clip->core.height;
	rep.canvas_width = child->core.width;
	rep.canvas_height = child->core.height;
	XtCallCallbackList ((Widget) w, w->linkview.report_callbacks,
			    (XtPointer) &rep);
    }
}


static void MoveChild(w, x, y)
    LinkViewWidget w;
    Position x, y;
{
    Widget child = w->linkview.child;
    Widget clip = w->linkview.clip;

    /* make sure we never move past right/bottom borders */
    if (-x + (int)clip->core.width > (int)child->core.width)
	x = -(child->core.width - clip->core.width);

    if (-y + (int)clip->core.height > (int)child->core.height)
	y = -(child->core.height - clip->core.height);

    /* make sure we never move past left/top borders */
    if (x >= 0) x = 0;
    if (y >= 0) y = 0;

    XtMoveWidget(child, x, y);
    SendReport (w, (XawPRSliderX | XawPRSliderY));

    RedrawThumbs(w);
}


static void ComputeLayout(widget, query, destroy_scrollbars)
    Widget widget;		/* LinkView */
    Boolean query;		/* query child's preferred geom? */
    Boolean destroy_scrollbars;	/* destroy un-needed scrollbars? */
{
    LinkViewWidget w = (LinkViewWidget)widget;
    Widget child = w->linkview.child;
    Widget clip = w->linkview.clip;
    LinkViewConstraints constraints
	= (LinkViewConstraints)clip->core.constraints;
    Boolean needshoriz, needsvert;
    int clip_width, clip_height;
    XtWidgetGeometry intended;

/*printf("LinkView: ComputeLayout(): query=%d\n", query);*/
    if (child == (Widget) NULL) return;

    clip_width = w->core.width;
    clip_height = w->core.height;
    intended.request_mode = CWBorderWidth;
    intended.border_width = 0;

    if (w->linkview.forcebars) {
        needsvert = w->linkview.allowvert;
        needshoriz = w->linkview.allowhoriz;
        ComputeWithForceBars(widget, query, &intended, 
			     &clip_width, &clip_height);
    }
    else {
        Dimension prev_width, prev_height;
		XtGeometryMask prev_mode;
		XtWidgetGeometry preferred;

		needshoriz = needsvert = False;

		/*
		 * intended.{width,height} caches the eventual child dimensions,
		 * but we don't set the mode bits until after we decide that the
		 * child's preferences are not acceptable.
		 */
/*printf("LinkView: ComputeLayout: cw=%d, ch=%d\n", clip_width, clip_height); */

		if (!w->linkview.allowhoriz) 
		    intended.request_mode |= CWWidth;

		if ((int)child->core.width < clip_width) 
		    intended.width = clip_width;
		else
		    intended.width = child->core.width;

		if ((int)child->core.height < clip_height) 
		    intended.height = clip_height;
		else
		    intended.height = child->core.height;

		if (!w->linkview.allowvert) 
		    intended.request_mode |= CWHeight;

		if (!query) {
		    preferred.width = child->core.width;
	    	preferred.height = child->core.height;
			}

		do { /* while intended != prev  */
		    if (query) {
	    	    (void) XtQueryGeometry( child, &intended, &preferred );
/*
printf("LinkView: ComputeLayout: iw=%d, ih=%d\n", intended.width, intended.height);
printf("LinkView: ComputeLayout: pw=%d, ph=%d\n", preferred.width, preferred.height);
*/
				if ( !(preferred.request_mode & CWWidth) )
				    preferred.width = intended.width;
				if ( !(preferred.request_mode & CWHeight) )
		    		preferred.height = intended.height;
		    	}
		    prev_width = intended.width;
		    prev_height = intended.height;
	    	prev_mode = intended.request_mode;

		    /*
		     * note that having once decided to turn on either bar
	    	 * we'll not change our mind until we're next resized,
		     * thus avoiding potential oscillations.
		     */

#define CheckHoriz()							  \
	    if ( w->linkview.allowhoriz && (int)preferred.width > clip_width) { 			  \
	        if (!needshoriz) {					  \
			    Widget bar;					          \
			    needshoriz = True;				          \
			    if ((bar = w->linkview.horiz_bar) == (Widget)NULL)    \
		    	    bar = CreateScrollbar(w, True);		          \
			    clip_height -= bar->core.height + bar->core.border_width; \
		    	if (clip_height < 1) clip_height = 1;		  \
				}							  \
	        intended.width = preferred.width;			  \
	    }
/*enddef*/

		    CheckHoriz();
		    if (w->linkview.allowvert && (int)preferred.height > clip_height) {
	    	    if (!needsvert) {
		    		Widget bar;
				    needsvert = True;
				    if ((bar = w->linkview.vert_bar) == (Widget)NULL)
		        	bar = CreateScrollbar(w, False);
				    clip_width -= bar->core.width + bar->core.border_width;
				    if (clip_width < 1) clip_width = 1;
				    CheckHoriz();
					}
				intended.height = preferred.height;
	    		}
		    if ( !w->linkview.allowhoriz ||
			 (int)preferred.width < clip_width) {
	    	    intended.width = clip_width;
				intended.request_mode |= CWWidth;
	    		}
		    if ( !w->linkview.allowvert ||
			 (int)preferred.height < clip_height) {
	    	    intended.height = clip_height;
				intended.request_mode |= CWHeight;
	    		}
		} while (intended.request_mode != prev_mode
		 || (intended.request_mode & CWWidth
		     && intended.width != prev_width)
		 || (intended.request_mode & CWHeight
		     && intended.height != prev_height));
    }

/*printf("    iw=%d, ih=%d\n", intended.width, intended.height);*/
    if (XtIsRealized(clip))
		XRaiseWindow( XtDisplay(clip), XtWindow(clip) );

    XtMoveWidget( clip,
		  (Position) (needsvert ? (w->linkview.useright ? 0 :
			       w->linkview.vert_bar->core.width +
			       w->linkview.vert_bar->core.border_width) : 0),
		  (Position) (needshoriz ? (w->linkview.usebottom ? 0 :
				w->linkview.horiz_bar->core.height +
			        w->linkview.horiz_bar->core.border_width) : 0));
    XtResizeWidget( clip, (Dimension)clip_width,
		    (Dimension)clip_height, (Dimension)0 );
	
    if (w->linkview.horiz_bar != (Widget)NULL) {
		Widget bar = w->linkview.horiz_bar;
		if (!needshoriz) {
	    	constraints->form.vert_base = (Widget)NULL;
		    if (destroy_scrollbars) {
				XtDestroyWidget( bar );
				w->linkview.horiz_bar = (Widget)NULL;
	    		}
			}
		else {
		    int bw = bar->core.border_width;
	    	XtResizeWidget( bar, (Dimension) clip_width, bar->core.height, (Dimension) bw );
		    XtMoveWidget( bar, (Position)((needsvert && !w->linkview.useright)
			   ? w->linkview.vert_bar->core.width
			   : -bw),
			  (Position)(w->linkview.usebottom
			    ? w->core.height - bar->core.height - bw
			    : -bw) );
		    XtSetMappedWhenManaged( bar, True );
			}
    	}

    if (w->linkview.vert_bar != (Widget)NULL) {
		Widget bar = w->linkview.vert_bar;
		if (!needsvert) {
	    	constraints->form.horiz_base = (Widget)NULL;
		    if (destroy_scrollbars) {
				XtDestroyWidget( bar );
				w->linkview.vert_bar = (Widget)NULL;
	    		}
			}
		else {
		    int bw = bar->core.border_width;
		    XtResizeWidget( bar, bar->core.width, (Dimension)clip_height, (Dimension)bw );
		    XtMoveWidget( bar,
			  (Position)(w->linkview.useright
			   ? w->core.width - bar->core.width - bw 
			   : -bw),
			  (Position)((needshoriz && !w->linkview.usebottom)
			    ? w->linkview.horiz_bar->core.height
			    : -bw) );
	    	XtSetMappedWhenManaged( bar, True );
			}
    	}

    if (child != (Widget)NULL) {
/*printf("LinkView: ComputeLayout: resizing child w=%d, h=%d\n", intended.width,	intended.height);*/
		XtResizeWidget( child, (Dimension)intended.width, (Dimension)intended.height, (Dimension)0 );
		MoveChild(w, needshoriz ? child->core.x : 0, needsvert ? child->core.y : 0);
    	}

    SendReport (w, XawPRAll);
	}



/*      Function Name: ComputeWithForceBars
 *      Description: Computes the layout give forcebars is set.
 *      Arguments: widget - the linkview widget.
 *                 query - whether or not to query the child.
 *                 intended - the cache of the childs height is
 *                            stored here ( USED AND RETURNED ).
 *                 clip_width, clip_height - size of clip window.
 *                                           (USED AND RETURNED ).
 *      Returns: none.
 */

static void
ComputeWithForceBars(widget, query, intended, clip_width, clip_height)
Widget widget;
Boolean query;
XtWidgetGeometry * intended;
int *clip_width, *clip_height;
{
    LinkViewWidget w = (LinkViewWidget)widget;
    Widget child = w->linkview.child;
    XtWidgetGeometry preferred;

/*
 * If forcebars then needs = allows = has.
 * Thus if needsvert is set it MUST have a scrollbar.
 */

    if (w->linkview.allowvert) {
	if (w->linkview.vert_bar == NULL) 
	    w->linkview.vert_bar = CreateScrollbar(w, False);

	*clip_width -= w->linkview.vert_bar->core.width +
		       w->linkview.vert_bar->core.border_width;
    }

    if (w->linkview.allowhoriz) {
	if (w->linkview.horiz_bar == NULL) 
	    w->linkview.horiz_bar = CreateScrollbar(w, True);

        *clip_height -= w->linkview.horiz_bar->core.height +
		       w->linkview.horiz_bar->core.border_width;
    }

    AssignMax( *clip_width, 1 );
    AssignMax( *clip_height, 1 );

    if (!w->linkview.allowvert) {
        intended->height = *clip_height;
        intended->request_mode = CWHeight;
    }
    if (!w->linkview.allowhoriz) {
        intended->width = *clip_width;
        intended->request_mode = CWWidth;
    }

    if ( query ) {
        if ( (w->linkview.allowvert || w->linkview.allowhoriz) ) { 
	    XtQueryGeometry( child, intended, &preferred );
	  
	    if ( !(intended->request_mode & CWWidth) )
	        if ( preferred.request_mode & CWWidth )
		    intended->width = preferred.width;
		else
		    intended->width = child->core.width;

	    if ( !(intended->request_mode & CWHeight) )
	        if ( preferred.request_mode & CWHeight )
		    intended->height = preferred.height;
		else
		    intended->height = child->core.height;
	}
    }
    else {
        if (w->linkview.allowvert)
	    intended->height = child->core.height;
	if (w->linkview.allowhoriz)
	    intended->width = child->core.width;
    }

    if (*clip_width > (int)intended->width)
	intended->width = *clip_width;
    if (*clip_height > (int)intended->height)
	intended->height = *clip_height;
}

static void Resize(widget)
    Widget widget;
{
	LinkViewWidget w = (LinkViewWidget) widget;
    Widget v_bar = w->linkview.vert_bar;
	Dimension width=0;

	if (v_bar != NULL)
		width = v_bar->core.width + v_bar->core.border_width;

	if (w->linkview.child_resizable) 
		XtVaSetValues(w->linkview.child, XtNwidth, w->core.width-width, NULL);
    ComputeLayout( widget, /*query=*/True, /*destroy=*/True );
}


/* ARGSUSED */
static Boolean Layout(w, width, height)
    FormWidget w;
    Dimension width, height;
{
    ComputeLayout( (Widget)w, /*query=*/True, /*destroy=*/True );
    w->form.preferred_width = w->core.width;
    w->form.preferred_height = w->core.height;
    return False;
}


static void ScrollUpDownProc(widget, closure, call_data)
    Widget widget;
    XtPointer closure;
    XtPointer call_data;
{
    LinkViewWidget w = (LinkViewWidget)closure;
    Widget child = w->linkview.child;
    int pix = (int)call_data;
    Position x, y;

    if (child == NULL) return;	/* no child to scroll. */

    x = child->core.x - ((widget == w->linkview.horiz_bar) ? pix : 0);
    y = child->core.y - ((widget == w->linkview.vert_bar) ? pix : 0);
    MoveChild(w, x, y);
}


/* ARGSUSED */
static void ThumbProc(widget, closure, call_data)
    Widget widget;
    XtPointer closure;
    XtPointer call_data;
{
    LinkViewWidget w = (LinkViewWidget)closure;
    Widget child = w->linkview.child;
    float *percent = (float *) call_data;
    Position x, y;

    if (child == NULL) return;	/* no child to scroll. */

    if (widget == w->linkview.horiz_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	x = (int)(-*percent * child->core.width);
#else /* else not macII */
	x = -(int)(*percent * child->core.width);
#endif /* macII */
    else
	x = child->core.x;

    if (widget == w->linkview.vert_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	y = (int)(-*percent * child->core.height);
#else /* else not macII */
	y = -(int)(*percent * child->core.height);
#endif /* macII */
    else
	y = child->core.y;

    MoveChild(w, x, y);
}

static XtGeometryResult
TestSmaller(w, request, reply_return)
     LinkViewWidget w; XtWidgetGeometry *request, *reply_return;
{
  if (request->width < w->core.width || request->height < w->core.height)
    return XtMakeGeometryRequest((Widget)w, request, reply_return);
  else
    return XtGeometryYes;  
}

static XtGeometryResult
GeometryRequestPlusScrollbar(w, horizontal, request, reply_return)
     Boolean horizontal;
     LinkViewWidget w; 
     XtWidgetGeometry *request, *reply_return;
{
  Widget sb;
  XtWidgetGeometry plusScrollbars;
  plusScrollbars = *request;
  if ((sb = w->linkview.horiz_bar) == (Widget)NULL)
    sb = CreateScrollbar( w, horizontal);
  request->width += sb->core.width;
  request->height += sb->core.height;
  XtDestroyWidget(sb);
  return XtMakeGeometryRequest((Widget) w, &plusScrollbars, reply_return);
 }

#define WidthChange() (request->width != w->core.width)
#define HeightChange() (request->height != w->core.height)

static XtGeometryResult 
QueryGeometry(w, request, reply_return)
     LinkViewWidget w; XtWidgetGeometry *request, *reply_return;
{	
  if (w->linkview.allowhoriz && w->linkview.allowvert) 
    return TestSmaller(w, request, reply_return);

  else if (w->linkview.allowhoriz && !w->linkview.allowvert) {
    if (WidthChange() && !HeightChange())
      return TestSmaller(w, request, reply_return);
    else if (!WidthChange() && HeightChange())
      return XtMakeGeometryRequest((Widget) w, request, reply_return);
    else if (WidthChange() && HeightChange()) /* hard part */
      return GeometryRequestPlusScrollbar(w, True, request, reply_return);
    else /* !WidthChange() && !HeightChange() */
      return XtGeometryYes;
  }
  else if (!w->linkview.allowhoriz && w->linkview.allowvert) {
    if (!WidthChange() && HeightChange())
      return TestSmaller(w, request, reply_return);
    else if (WidthChange() && !HeightChange())
      return XtMakeGeometryRequest((Widget)w, request, reply_return);
    else if (WidthChange() && HeightChange()) /* hard part */
      return GeometryRequestPlusScrollbar(w, False, request, reply_return);
    else /* !WidthChange() && !HeightChange() */
      return XtGeometryYes;
  }      
  else /* (!w->linkview.allowhoriz && !w->linkview.allowvert) */
    return XtMakeGeometryRequest((Widget) w, request, reply_return);
}

#undef WidthChange
#undef HeightChange

static XtGeometryResult GeometryManager(child, request, reply)
    Widget child;
    XtWidgetGeometry *request, *reply;
{
    LinkViewWidget w = (LinkViewWidget)child->core.parent;
    Boolean rWidth = (Boolean)(request->request_mode & CWWidth);
    Boolean rHeight = (Boolean)(request->request_mode & CWHeight);
    XtWidgetGeometry allowed;
    XtGeometryResult result;
    Boolean reconfigured;
    Boolean child_changed_size;
    Dimension height_remaining;

/*printf("LinkView: GeometryManager:\n");*/
    if (request->request_mode & XtCWQueryOnly)
      return QueryGeometry(w, request, reply);

    if (child != w->linkview.child
        || request->request_mode & ~(CWWidth | CWHeight
				     | CWBorderWidth)
	|| ((request->request_mode & CWBorderWidth)
	    && request->border_width > 0))
	return XtGeometryNo;

/*printf("LinkView: GeometryManager: rw=%d, rh=%d\n", request->width, request->height);*/
    allowed = *request;

    reconfigured = GetGeometry( (Widget)w,
			        (rWidth ? request->width : w->core.width),
			        (rHeight ? request->height : w->core.height)
			      );

    child_changed_size = ((rWidth && child->core.width != request->width) ||
			  (rHeight && child->core.height != request->height));

    height_remaining = w->core.height;
    if (rWidth && w->core.width != request->width) {
	if (w->linkview.allowhoriz && request->width > w->core.width) {
	    /* horizontal scrollbar will be needed so possibly reduce height */
	    Widget bar; 
	    if ((bar = w->linkview.horiz_bar) == (Widget)NULL)
		bar = CreateScrollbar( w, True );
	    height_remaining -= bar->core.height + bar->core.border_width;
	    reconfigured = True;
	}
	else {
	    allowed.width = w->core.width;
	}
    }
    if (rHeight && height_remaining != request->height) {
	if (w->linkview.allowvert && request->height > height_remaining) {
	    /* vertical scrollbar will be needed, so possibly reduce width */
	    if (!w->linkview.allowhoriz || request->width < w->core.width) {
		Widget bar;
		if ((bar = w->linkview.vert_bar) == (Widget)NULL)
		    bar = CreateScrollbar( w, False );
		if (!rWidth) {
		    allowed.width = w->core.width;
		    allowed.request_mode |= CWWidth;
		}
		if ( (int)allowed.width >
		     (int)(bar->core.width + bar->core.border_width) )
		    allowed.width -= bar->core.width + bar->core.border_width;
		else
		    allowed.width = 1;
		reconfigured = True;
	    }
	}
	else {
	    allowed.height = height_remaining;
	}
    }

    if (allowed.width != request->width || allowed.height != request->height) {
	*reply = allowed;
	result = XtGeometryAlmost;
    }
    else {
	if (rWidth)  child->core.width = request->width;
	if (rHeight) child->core.height = request->height;
	result = XtGeometryYes;
    }

    if (reconfigured || child_changed_size)
	ComputeLayout( (Widget)w,
		       /*query=*/ False,
		       /*destroy=*/ (result == XtGeometryYes) ? True : False );

    return result;
  }


static Boolean GetGeometry(w, width, height)
    Widget w;
    Dimension width, height;
{
    XtWidgetGeometry geometry, return_geom;
    XtGeometryResult result;

    if (width == w->core.width && height == w->core.height)
	return False;

    geometry.request_mode = CWWidth | CWHeight;
    geometry.width = width;
    geometry.height = height;

    if (XtIsRealized(w)) {
	if (((LinkViewWidget)w)->linkview.allowhoriz && width > w->core.width)
	    geometry.width = w->core.width;
	if (((LinkViewWidget)w)->linkview.allowvert && height > w->core.height)
	    geometry.height = w->core.height;
    } else {
	/* This is the Realize call; we'll inherit a w&h iff none currently */
	if (w->core.width != 0) {
	    if (w->core.height != 0) return False;
	    geometry.width = w->core.width;
	}
	if (w->core.height != 0) geometry.height = w->core.height;
    }

    result = XtMakeGeometryRequest(w, &geometry, &return_geom);
    if (result == XtGeometryAlmost)
	result = XtMakeGeometryRequest(w, &return_geom, (XtWidgetGeometry *)NULL);

    return (result == XtGeometryYes);
}

static XtGeometryResult PreferredGeometry(w, constraints, reply)
    Widget w;
    XtWidgetGeometry *constraints, *reply;
{
    if (((LinkViewWidget)w)->linkview.child != NULL)
	return XtQueryGeometry( ((LinkViewWidget)w)->linkview.child,
			       constraints, reply );
    else
	return XtGeometryYes;
}


void
#if NeedFunctionPrototypes
LinkViewReload(Widget gw,
#if NeedWidePrototypes
			   int x, int y)
#else
			   Position x, Position y)
#endif
#else
LinkViewReload(gw, x, y)
    Widget gw;
    Position x, y;
#endif
{
    LinkViewWidget w = (LinkViewWidget) gw;
    Widget child = w->linkview.child;

	ComputeLayout(gw, True, True);
    if (x > (int)child->core.width) 
      x = child->core.width;
    else if (x < 0)
      x = child->core.x;

    if (y > (int)child->core.height)
      y = child->core.height;
    else if (y < 0)
      y = child->core.y;

    MoveChild (w, -x, -y);
}

void
#if NeedFunctionPrototypes
LinkViewGetCoordinates(Widget gw, int *x, int *y)
#else
LinkViewGetCoordinates(gw, x, y)
    Widget gw;
    int *x, *y;
#endif
{
    LinkViewWidget w = (LinkViewWidget) gw;
    Widget child = w->linkview.child;

	*x = - child->core.x;
	*y = - child->core.y;
}

void
#if NeedFunctionPrototypes
LinkViewSetCoordinates (Widget gw,
#if NeedWidePrototypes
                           int x, int y)
#else
                           Position x, Position y)
#endif
#else
LinkViewSetCoordinates (gw, x, y)
    Widget gw;
    Position x, y;
#endif
{
    LinkViewWidget w = (LinkViewWidget) gw;
    Widget child = w->linkview.child;

    if (x > (int)child->core.width) 
      x = child->core.width;
    else if (x < 0)
      x = child->core.x;

    if (y > (int)child->core.height)
      y = child->core.height;
    else if (y < 0)
      y = child->core.y;

    MoveChild (w, -x, -y);
}


/* Action procedures */

static void lineDown(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	MoveChild(w, child->core.x, child->core.y - w->linkview.baselineskip);
	}

static void lineUp(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	MoveChild(w, child->core.x, child->core.y + w->linkview.baselineskip);
	}

static void pagePrev(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	if (XtClass(child) == linkTextWidgetClass)
		LinkTextPrevPage(child, True);
	else
		XtWarning("The pagePrev action is appropriate only for LinkText child widgets.");
	}

static void pageNext(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	if (XtClass(child) == linkTextWidgetClass)
		LinkTextNextPage(child, True);
	else
		XtWarning("The pageNext action is appropriate only for LinkText child widgets.");
	}

static void pageFirst(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	if (XtClass(child) == linkTextWidgetClass)
		LinkTextFirstPage(child, True);
	else
		XtWarning("The pageFirst action is appropriate only for LinkText child widgets.");
	}

static void pageLast(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	if (XtClass(child) == linkTextWidgetClass)
		LinkTextLastPage(child, False);
	else
		XtWarning("The pageLast action is appropriate only for LinkText child widgets.");
	}

static void pageDown(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	if (child->core.y == -(child->core.height - clip->core.height)) {
		if (XtClass(child) == linkTextWidgetClass) 
			LinkTextNextPage(child, True);
		}
	else
		MoveChild(w, child->core.x, child->core.y - clip->core.height);
	}

static void pageUp(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	if (child->core.y == 0) {
		if (XtClass(child) == linkTextWidgetClass) 
			LinkTextPrevPage(child, False);
		}
	else
		MoveChild(w, child->core.x, child->core.y + clip->core.height);
	}

static void pageTop(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;

	MoveChild(w, child->core.x, 0);
	}

static void pageBottom(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	MoveChild(w, child->core.x, -child->core.height + clip->core.height);
	}

static void pageLeft(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	MoveChild(w, child->core.x + (clip->core.width/2), child->core.y);
	}

static void pageRight(Widget widget, XEvent *ev, String *s, Cardinal *num) {
	LinkViewWidget w = (LinkViewWidget) widget;
	Widget child = w->linkview.child;
	Widget clip = w->linkview.clip;

	MoveChild(w, child->core.x - (clip->core.width/2), child->core.y);
	}

