# PcBar - progress bar widget.

@class PcBar (Core)     @file=PcBar


@public

@ The foreground colour.

	@var Pixel foreground = <String> XtDefaultForeground

@ The percentage value to be displayed.

	@var int percentage = 0

@ Whether to display a percentage in text.

	@var Boolean displaypc = False

@ If displaying percentage as text, whether to display a 0 value.

	@var Boolean showzero = False

@ If true, the bar will be vertical rather than horizontal.

	@var Boolean vertical = False

@ The font to display the text in.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@PRIVATE

@ The GC for drawing.

        @var GC draw_gc

@ The GC for writing.

        @var GC text_gc

@utilities

@ Initialize the GC for drawing.

@proc GetGC($) {
    XGCValues values;
    XtGCMask mask = GCForeground | GCBackground;

    values.foreground = $foreground;
    values.background = $background_pixel;

    $draw_gc = XtGetGC($, mask, &values);

	values.function = GXxor;
	values.foreground = $foreground ^ $background_pixel;
	mask = GCFunction | GCForeground | GCBackground;
    $text_gc = XtGetGC($, mask, &values);
	XSetFont(XtDisplay($), $text_gc, $font->fid); 
	}

@METHODS

@def H_DEFAULTWIDTH = 200  
@def H_DEFAULTHEIGHT = 20  
@def V_DEFAULTWIDTH = 30
@def V_DEFAULTHEIGHT = 200  

@proc initialize {

    if (($percentage < 0) ||	($percentage > 100))  {
		XtWarning("PcBar: invalid percentage (using 0)."); 
    	$percentage = 0;
	    }

    GetGC($);
    if ($width == 0) 
		$width = $vertical ? V_DEFAULTWIDTH : H_DEFAULTWIDTH;
    if ($height == 0) 
		$height = $vertical ? V_DEFAULTHEIGHT : H_DEFAULTHEIGHT;

	}

@proc expose {

    if (!XtIsRealized($))
		return;

	PcBarSetPercentage($, $percentage);
	}


@proc set_values {
    Boolean do_redisplay = False;

    if (($old$foreground != $foreground) ||
    	($old$background_pixel != $background_pixel) ||
     	($old$font != $font)) {
		XtReleaseGC($, $draw_gc);
		XtReleaseGC($, $text_gc);
		GetGC($);
		do_redisplay = True;
	    }
    if ($old$percentage != $percentage) 
		do_redisplay = True;
    if ($old$displaypc != $displaypc) 
		do_redisplay = True;
    if (($old$showzero != $showzero) && ($percentage == 0)) {
		do_redisplay = True;
		}
    if ($old$vertical != $vertical) 
		do_redisplay = True;

    return do_redisplay;
	}


@proc destroy {
    if ($draw_gc)
		XtReleaseGC($, $draw_gc);
    if ($text_gc)
		XtReleaseGC($, $text_gc);
	}



@proc query_geometry {
	/* accept any requested change in size */
    return XtGeometryYes;
	}

@EXPORTS

@proc void PcBarSetPercentage($, int pc) {
    Display *disp;
    Window win;
	int x,y;
	int tx,ty;
	char buf[10];
	int asc, des, dir;
	XCharStruct overall;

	$percentage=pc;
    if (!XtIsRealized($))
		return;

	disp=XtDisplay($);
	win=XtWindow($);

    XSetLineAttributes(disp, $draw_gc, $height, LineSolid,CapButt,JoinBevel);
	if ($vertical) {
	    XSetLineAttributes(disp, $draw_gc, $width, LineSolid,CapButt,JoinBevel);
	    y= $height - ($percentage * $height)/100;
	    XDrawLine(disp, win,$draw_gc,  $width/2, y, $width/2, $height);
    	if ($percentage <100) 
			XClearArea(disp, win, 0, 0, $width, y, False);
		}
	else {
	    XSetLineAttributes(disp,$draw_gc, $height, LineSolid,CapButt,JoinBevel);
	    x=($percentage * $width)/100;
	    XDrawLine(disp, win, $draw_gc, 0, $height/2, x, $height/2);
    	XClearArea(disp, win, x, 0, $width, $height, False);
		}

	if (($displaypc) && ($showzero || $percentage > 0)) {
		sprintf(buf, "%d%%", $percentage);
        XTextExtents($font, buf, strlen(buf), &dir, &asc, &des, &overall);
		tx = ($width-overall.width)/2;
		ty = ($height - (asc+des))/2 + asc;
		XDrawString(disp, win, $text_gc, tx, ty,  buf, strlen(buf));
		}
	}

@imports

@incl <stdio.h>
